"""QuartzBio Object API resource"""

import os
import re
import base64
import binascii
import mimetypes
import sys
import time
from datetime import datetime

import requests
from urllib3.util.retry import Retry

from quartzbio.errors import QuartzBioError
from quartzbio.errors import NotFoundError
from quartzbio.errors import FileUploadError
from quartzbio.utils.md5sum import md5sum
from quartzbio.utils.files import separate_filename_extension

from ..client import client

from .quartzbio_object import convert_to_quartzbio_object

from .apiresource import CreateableAPIResource
from .apiresource import ListableAPIResource
from .apiresource import SearchableAPIResource
from .apiresource import UpdateableAPIResource
from .apiresource import DeletableAPIResource
from .apiresource import DownloadableAPIResource


class UploadProgressTracker:
    """Simple progress tracking for multipart uploads."""

    def __init__(self, total_parts, total_size):
        self.total_parts = total_parts
        self.total_size = total_size
        self.completed_parts = 0
        self.completed_bytes = 0
        self.start_time = time.time()
        self.progress_line_active = False

    def update_progress(self, part_size, part_duration=None):
        """Update progress with new part completion - overwrites same line"""
        self.completed_parts += 1
        self.completed_bytes += part_size

        # Calculate metrics
        elapsed = time.time() - self.start_time
        avg_speed = self.completed_bytes / elapsed if elapsed > 0 else 0
        remaining_bytes = self.total_size - self.completed_bytes
        eta = remaining_bytes / avg_speed if avg_speed > 0 else 0

        # Format bytes for display
        completed_mb = self.completed_bytes / 1024 / 1024
        total_mb = self.total_size / 1024 / 1024
        speed_mb = avg_speed / 1024 / 1024

        # Create simple progress message
        progress_msg = f"{self.completed_parts}/{self.total_parts} parts uploaded"
        progress_msg += f" ({completed_mb:.1f}MB/{total_mb:.1f}MB)"

        if avg_speed > 0:
            progress_msg += f" [{speed_mb:.1f}MB/s"
            if eta > 0:
                progress_msg += f", ETA: {eta:.0f}s"
            progress_msg += "]"

        # Use \r to overwrite the same line (like Unix downloads)
        print(f"\r{progress_msg}", end="", flush=True)
        self.progress_line_active = True

        # Print newline only when complete
        if self.completed_parts >= self.total_parts:
            print()  # Final newline
            self.progress_line_active = False

    def get_completion_percentage(self):
        """Get completion percentage"""
        return (
            (self.completed_bytes / self.total_size) * 100 if self.total_size > 0 else 0
        )

    def get_elapsed_time(self):
        """Get total elapsed time"""
        return time.time() - self.start_time

    def get_average_speed(self):
        """Get average upload speed in bytes/second"""
        elapsed = time.time() - self.start_time
        return self.completed_bytes / elapsed if elapsed > 0 else 0

    def notify_error(self):
        """Notify that an error message will be printed - move to new line"""
        if self.progress_line_active:
            print()  # Move to new line before error message
            self.progress_line_active = False


if sys.version_info >= (3, 9, 0):
    from collections.abc import Iterable
else:
    from collections import Iterable  # noqa


class Object(
    CreateableAPIResource,
    ListableAPIResource,
    DeletableAPIResource,
    SearchableAPIResource,
    UpdateableAPIResource,
    DownloadableAPIResource,
):
    """
    An object is a resource in a Vault.  It has three possible types,
    though more may be added later: folder, file, and QuartzBio Dataset.
    """

    RESOURCE_VERSION = 2

    LIST_FIELDS = (
        ("id", "ID"),
        ("object_type", "Type"),
        ("full_path", "Full Path"),
        ("description", "Description"),
    )

    # Regex describing an object path.
    PATH_RE = re.compile(r"^[^\/]*(?P<path>(\/[^\/]*)+)$")

    @classmethod
    def validate_full_path(cls, full_path, **kwargs):
        """Helper method to parse a full or partial path and
        return a full path as well as a dict containing path parts.

        Uses the following rules when processing the path:

            * If no domain, uses the current user's account domain
            * If no vault, uses the current user's personal vault.
            * If no path, uses '/' (vault root)

        Returns a tuple containing:

            * The validated full_path
            * A dictionary with the components:
                * domain: the domain of the vault
                * vault: the name of the vault, without domain
                * vault_full_path: domain:vault
                * path: the object path within the vault
                * parent_path: the parent path to the object
                * parent_full_path: the parent full path to the object
                * filename: the object's filename (if any)
                * full_path: the validated full path

        The following components may be overridden using kwargs:

            * vault
            * path

        Object paths (also known as "paths") must begin with a forward slash.

        The following path formats are supported:

            domain:vault:/path -> object "path" in the root of "domain:vault"
            domain:vault/path  -> object "path" in the root of "domain:vault"
            vault:/path        -> object "path" in the root of "vault"
            vault/path         -> object "path" in the root of "vault"
            ~/path             -> object "path" in the root of personal vault
            vault/             -> root of "vault"
            ~/                 -> root of your personal vault

        The following two formats are not supported:

            path               -> invalid/ambiguous path (exception)
            vault:path         -> invalid/ambiguous path (exception)
            vault:path/path    -> unsupported, interpreted as domain:vault/path

        """
        from quartzbio.resource.vault import Vault

        _client = kwargs.pop("client", None) or cls._client or client

        if not full_path:
            raise Exception(
                "Invalid path: ",
                "Full path must be in one of the following formats: "
                '"vault:/path", "domain:vault:/path", or "~/path"',
            )

        # Parse the vault's full_path, using overrides if any
        input_vault = kwargs.get("vault") or full_path
        try:
            vault_full_path, path_dict = Vault.validate_full_path(
                input_vault, client=_client
            )
        except Exception as err:
            raise Exception(
                'Could not determine vault from "{0}": {1}'.format(input_vault, err)
            )

        if kwargs.get("path"):
            # Allow override of the object_path.
            full_path = "{0}:/{1}".format(vault_full_path, kwargs["path"])

        match = cls.PATH_RE.match(full_path)
        if match:
            object_path = match.groupdict()["path"]
        else:
            raise Exception(
                'Cannot find a valid object path in "{0}". '
                "Full path must be in one of the following formats: "
                '"vault:/path", "domain:vault:/path", or "~/path"'.format(full_path)
            )

        # Remove double slashes
        object_path = re.sub("//+", "/", object_path)
        if object_path != "/":
            # Remove trailing slash
            object_path = object_path.rstrip("/")

        path_dict["path"] = object_path
        path_dict["full_path"] = "{domain}:{vault}:{path}".format(**path_dict)

        # Assumes no trailing slash, will be '' if is a vault root
        path_dict["filename"] = os.path.basename(object_path)

        # Will be / if parent is vault root
        path_dict["parent_path"] = os.path.dirname(object_path)
        path_dict["parent_full_path"] = "{vault_full_path}:{parent_path}".format(
            **path_dict
        )

        return path_dict["full_path"], path_dict

    @classmethod
    def get_by_path(cls, path, **params):
        assert_type = params.pop("assert_type", None)
        params.update({"path": path})
        obj = cls._retrieve_helper("object", "path", path, **params)
        if obj and assert_type and obj["object_type"] != assert_type:
            raise QuartzBioError(
                "Expected a {} but found a {} at {}".format(
                    assert_type, obj["object_type"], path
                )
            )

        return obj

    @classmethod
    def get_by_full_path(cls, full_path, **params):
        # Don't pop client from params since **params is used below
        _client = params.get("client", None) or cls._client or client
        full_path, _ = cls.validate_full_path(full_path, client=_client)
        assert_type = params.pop("assert_type", None)
        params.update({"full_path": full_path})
        obj = cls._retrieve_helper("object", "full_path", full_path, **params)
        if obj and assert_type and obj["object_type"] != assert_type:
            raise QuartzBioError(
                "Expected a {} but found a {} at {}".format(
                    assert_type, obj["object_type"], full_path
                )
            )

        return obj

    @classmethod
    def get_or_create_by_full_path(cls, full_path, **kwargs):
        from quartzbio import Vault
        from quartzbio import Object

        _client = kwargs.pop("client", None) or cls._client or client
        create_vault = kwargs.pop("create_vault", False)
        create_folders = kwargs.pop("create_folders", True)

        # Check for object type assertion, if not explicitly added, see
        # if user has passed object_type, as their intent was to get/create
        # an object of that type.
        assert_type = kwargs.pop("assert_type", kwargs.get("object_type", None))

        try:
            return cls.get_by_full_path(
                full_path, assert_type=assert_type, client=_client
            )
        except NotFoundError:
            pass

        # Object type required when creating Object
        object_type = kwargs.get("object_type")
        if not object_type:
            raise Exception(
                "'object_type' is required when creating a new "
                "Object. Pass one of: file, folder, dataset"
            )

        # TODO should we require file contents?
        # Technically a user could then use this object to the call
        # upload_file()
        # if object_type == 'file' and not kwargs.get('content'):
        #     raise Exception('')

        # Object not found, create it step-by-step
        full_path, parts = Object.validate_full_path(full_path, client=_client)

        if create_vault:
            vault = Vault.get_or_create_by_full_path(
                "{0}:{1}".format(parts["domain"], parts["vault"]), client=_client
            )
        else:
            vaults = Vault.all(
                account_domain=parts["domain"], name=parts["vault"], client=_client
            )
            if len(vaults.quartzbio_objects()) == 0:
                raise Exception(
                    "Vault with name {0}:{1} does not exist. Pass "
                    "create_vault=True to auto-create".format(
                        parts["domain"], parts["vault"]
                    )
                )
            vault = vaults.quartzbio_objects()[0]

        # Create the folders to hold the object if they do not already exist.
        object_path = parts["path"]
        curr_path = os.path.dirname(object_path)
        folders_to_create = []
        new_folders = []
        id_map = {"/": None}

        while curr_path != "/":
            try:
                obj = Object.get_by_path(
                    curr_path, vault_id=vault.id, assert_type="folder", client=_client
                )
                id_map[curr_path] = obj.id
                break
            except NotFoundError:
                if not create_folders:
                    raise Exception(
                        "Folder {} does not exist.  Pass "
                        "create_folders=True to auto-create "
                        "missing folders"
                    )

                folders_to_create.append(curr_path)
                curr_path = "/".join(curr_path.split("/")[:-1])
                if curr_path == "":
                    break

        for folder in reversed(folders_to_create):
            new_folder = Object.create(
                object_type="folder",
                vault_id=vault.id,
                filename=os.path.basename(folder),
                parent_object_id=id_map[os.path.dirname(folder)],
                client=_client,
            )
            new_folders.append(new_folder)
            id_map[folder] = new_folder.id

        if os.path.dirname(object_path) == "/":
            parent_folder_id = None
        elif new_folders:
            parent_folder_id = new_folders[-1].id
        else:
            parent_folder_id = id_map[os.path.dirname(object_path)]

        return Object.create(
            filename=os.path.basename(object_path),
            vault_id=vault.id,
            parent_object_id=parent_folder_id,
            client=_client,
            **kwargs,
        )

    @staticmethod
    def _get_timestamp(date_format="%Y-%m-%d_%Hh%Mm%Ss_%Z"):
        return datetime.now().strftime(date_format)

    def _archive(self, archive_folder):
        if not self.object_type == "file":
            raise NotImplementedError(
                "Object archiving is only supported for files, not {}".format(
                    self.object_type
                )
            )

        # Create timestamped archive filename
        timestamp = self._get_timestamp()

        base_filename, file_extension, compression = separate_filename_extension(
            self.filename
        )
        archive_filename = "{base_filename}_{timestamp}{extension}{compression}".format(
            base_filename=base_filename,
            timestamp=timestamp,
            extension=file_extension,
            compression=compression,
        )

        # Create parent archive nested directory paths
        archive_path = os.path.join(
            archive_folder, os.path.dirname(self.path).lstrip("/"), archive_filename
        )

        # Ensure no errors with archive full path
        Object.validate_full_path(self.vault.full_path + ":/" + archive_path)

        # Create all parent folders if they don't already exist
        archive_parent_folder = os.path.dirname(archive_path).lstrip("/")
        if archive_parent_folder != "":
            folders = archive_parent_folder.split("/")
            parent_folder_path = self.vault.full_path + ":"
            for folder in folders:
                folder_full_path = os.path.join(parent_folder_path, folder)
                parent_folder = Object.create_folder(self.vault, folder_full_path)
                parent_folder_path = parent_folder.full_path
            self.parent_object_id = parent_folder.id
        else:
            self.parent_object_id = None

        print("Archiving file {} to {}".format(self.full_path, archive_path))
        # Multiple saves are needed as parent object ID
        # and filename cannot be updated in the same API call
        self.save()
        self.filename = archive_filename
        return self.save()

    @classmethod
    def create_folder(cls, vault, full_path, tags=None, **kwargs):
        """Create a folder if not exists.

        Args:
            vault (:class:`~quartzbio.resources.vault.Vault`): A Vault object.
            full_path (str): Full path including vault name.
            tags (list[str]): List of tags to put on folder.
            client: QuartzBio client configuration to use.
        Returns:
            Object: New folder object
        Raises:
            :class:`~quartzbio.errors.QuartzBioError`: if a file or dataset object already exists
                at the given full_path.
        """
        _client = kwargs.pop("client", None) or cls._client or client

        full_path, path_dict = Object.validate_full_path(full_path, client=_client)
        folder_name = path_dict["filename"]

        try:
            new_obj = Object.get_by_full_path(full_path, client=_client)
            if not new_obj.is_folder:
                raise QuartzBioError(
                    "Object type {} already exists at location: {}".format(
                        new_obj.object_type, full_path
                    )
                )
        except NotFoundError:
            # Create the folder
            if path_dict["parent_path"] == "/":
                parent_object_id = None
            else:
                parent = Object.get_by_full_path(
                    path_dict["parent_full_path"], assert_type="folder", client=_client
                )
                parent_object_id = parent.id

            # Make the API call
            print("Creating folder {}".format(full_path))
            new_obj = Object.create(
                vault_id=vault.id,
                parent_object_id=parent_object_id,
                object_type="folder",
                filename=folder_name,
                tags=tags or [],
                client=_client,
            )

            print(
                "Notice: Folder created for {0} at {1}".format(
                    folder_name, new_obj.path
                )
            )

        return new_obj

    def create_shortcut(self, shortcut_full_path, **kwargs):
        """Create a shortcut to the current object at shortcut_full_path

        Args:
            shortcut_full_path (str): Full path including vault name.
            tags (list[str]): List of tags to put on shortcut.
        Returns:
            Object: New shortcut object
        Raises:
            :class:`~quartzbio.errors.QuartzBioError`: if a object already exists at the given shortcut_full_path.
        """
        _client = kwargs.pop("client", None) or self._client or client
        full_path, path_dict = Object.validate_full_path(
            shortcut_full_path, client=_client
        )

        try:
            existing_object = Object.get_by_full_path(full_path, client=_client)
            if existing_object:
                raise QuartzBioError(
                    "Object type {} already exists at location: {}".format(
                        existing_object.object_type, full_path
                    )
                )
        except NotFoundError:
            target = {"id": self.id, "object_type": self.object_type}

            shortcut = Object.get_or_create_by_full_path(
                full_path=shortcut_full_path,
                object_type="shortcut",
                target=target,
                client=_client,
            )

            print("Notice: Shortcut created at {}".format(shortcut.path))

        return shortcut

    @classmethod
    def upload_file(cls, local_path, remote_path, vault_full_path, **kwargs):
        """Upload a file to a QuartzBio vault.

        Automatically uses multipart upload for files larger than the multipart_threshold.

        Args:
            local_path (str): Path to the local file to upload
            remote_path (str): Remote path within the vault
            vault_full_path (str): Full path of the target vault
            **kwargs: Additional options:
                multipart_threshold (int): File size threshold for multipart upload (default: 64MB)
                multipart_chunksize (int): Size of each upload part (default: 64MB)
                num_processes (int): Number of parallel workers for multipart upload (default: 1)
                max_retries (int): Maximum retries per part for multipart upload (default: 3)
                archive_folder (str): Archive existing files to this folder before upload
                follow_shortcuts (bool): Follow shortcuts when uploading
                description (str): Description for the uploaded file
                tags (list): Tags to apply to the uploaded file
                client: QuartzBio client instance to use

        Returns:
            Object: The uploaded file object

        Raises:
            FileUploadError: If upload fails
            QuartzBioError: If API request fails
        """
        from quartzbio import Vault
        from quartzbio import Object

        _client = kwargs.pop("client", None) or cls._client or client

        local_path = os.path.expanduser(local_path)

        # Get vault
        vault = Vault.get_by_full_path(vault_full_path, client=_client)

        # Get MD5 and check if multipart upload is needed
        multipart_threshold = kwargs.get(
            "multipart_threshold", 64 * 1024 * 1024
        )  # 64MB default
        local_md5, _ = md5sum(local_path, multipart_threshold=multipart_threshold)
        # Get a mimetype of file
        mime_tuple = mimetypes.guess_type(local_path)
        # If a file is compressed get a compression type, otherwise a file type
        mimetype = mime_tuple[1] if mime_tuple[1] else mime_tuple[0]
        # Get file size
        size = os.path.getsize(local_path)
        if size == 0:
            print("WARNING: skipping empty object: {}".format(local_path))
            return False

        # Check if object exists already and compare md5sums
        full_path, path_dict = Object.validate_full_path(
            os.path.join(
                "{}:{}".format(vault.full_path, remote_path),
                os.path.basename(local_path),
            ),
            client=_client,
        )
        try:
            obj = cls.get_by_full_path(full_path, client=_client)
            if not obj.is_file:
                print(
                    "WARNING: A {} currently exists at {}".format(
                        obj.object_type, full_path
                    )
                )
            else:
                # Check against md5sum of remote file
                if obj.md5 == local_md5:
                    print(
                        "WARNING: File {} (md5sum {}) already exists, "
                        "not uploading".format(full_path, local_md5)
                    )
                    return obj
                else:
                    if kwargs.get("archive_folder"):
                        obj._archive(kwargs["archive_folder"])
                    else:
                        print(
                            "WARNING: File {} exists on QuartzBio with different "
                            "md5sum (local: {} vs remote: {}) Uploading anyway, "
                            "but not overwriting.".format(full_path, local_md5, obj.md5)
                        )
        except NotFoundError:
            obj = None
            pass

        # Lookup parent object
        if kwargs.get("follow_shortcuts") and obj and obj.is_file:
            vault_id = obj.vault_id
            parent_object_id = obj.parent_object_id
            filename = obj.filename
        else:
            vault_id = vault.id
            filename = os.path.basename(local_path)
            if path_dict["parent_path"] == "/":
                parent_object_id = None
            else:
                parent_obj = Object.get_by_full_path(
                    path_dict["parent_full_path"], assert_type="folder", client=_client
                )
                parent_object_id = parent_obj.id

        description = kwargs.get("description")

        # Create the file, and upload it to the Upload URL
        obj = Object.create(
            vault_id=vault_id,
            parent_object_id=parent_object_id,
            object_type="file",
            filename=filename,
            md5=local_md5,
            mimetype=mimetype,
            size=size,
            description=description,
            tags=kwargs.get("tags", []) or [],
            client=_client,
        )
        print("Notice: File created for {0} at {1}".format(local_path, obj.path))
        print("Notice: Upload initialized")

        # Check if multipart upload is needed
        if hasattr(obj, "is_multipart") and obj.is_multipart:
            return cls._upload_multipart(obj, local_path, local_md5, **kwargs)
        else:
            return cls._upload_single_file(obj, local_path, **kwargs)

    @classmethod
    def _upload_single_file(cls, obj, local_path, **kwargs):
        """Handle single-part upload for smaller files"""
        import mimetypes

        # Get a mimetype of file
        mime_tuple = mimetypes.guess_type(local_path)
        # If a file is compressed get a compression type, otherwise a file type
        mimetype = mime_tuple[1] if mime_tuple[1] else mime_tuple[0]
        # Get file size
        size = os.path.getsize(local_path)

        # Get MD5 for single part upload
        local_md5, _ = md5sum(local_path, multipart_threshold=None)

        upload_url = obj.upload_url

        headers = {
            "Content-MD5": base64.b64encode(binascii.unhexlify(local_md5)),
            "Content-Type": mimetype,
            "Content-Length": str(size),
        }

        # Use a session with a retry policy to handle connection errors.
        session = requests.Session()
        max_retries = 5
        retry = Retry(
            total=max_retries,
            read=max_retries,
            connect=max_retries,
            backoff_factor=2,
            status_forcelist=(500, 502, 503, 504, 400),
            allowed_methods=["HEAD", "OPTIONS", "GET", "PUT", "POST"],
        )
        session.mount("https://", requests.adapters.HTTPAdapter(max_retries=retry))

        # Handle retries when upload fails due to an exception such as SSLError
        n_retries = 0
        while True:
            try:
                upload_resp = session.put(
                    upload_url, data=open(local_path, "rb"), headers=headers
                )
            except Exception as e:
                if n_retries == max_retries:
                    obj.delete(force=True)
                    raise FileUploadError(str(e))

                n_retries += 1
                print(
                    "WARNING: Retrying ({}/{}) failed upload for {}: {}".format(
                        n_retries, max_retries, local_path, e
                    )
                )
                time.sleep(2 * n_retries)
            else:
                break

        if upload_resp.status_code != 200:
            print(
                "WARNING: Upload status code for {0} was {1}".format(
                    local_path, upload_resp.status_code
                )
            )
            # Clean up the failed upload
            obj.delete(force=True)
            raise FileUploadError(upload_resp.content)
        else:
            print(
                "Notice: Successfully uploaded {0} to {1}".format(local_path, obj.path)
            )

        return obj

    @classmethod
    def refresh_presigned_urls(cls, upload_id, key, total_size, part_numbers, **kwargs):
        """Refresh presigned URLs for multipart upload

        Args:
            upload_id (str): The upload ID from the multipart upload
            key (str): The upload key/identifier
            total_size (int): Total size of the file being uploaded
            part_numbers (list[int]): List of part numbers to refresh URLs for
            **kwargs: Additional parameters including client

        Returns:
            list: List of presigned URL objects with part information
        """
        _client = kwargs.get("client") or cls._client or client

        payload = {
            "upload_id": upload_id,
            "key": key,
            "total_size": total_size,
            "part_numbers": part_numbers,
        }

        try:
            response = _client.post("/v2/presigned_urls", payload)

            if "presigned_urls" in response:
                return response["presigned_urls"]
            else:
                raise FileUploadError(
                    "Invalid response from presigned URLs API: missing 'presigned_urls' key"
                )
        except Exception as e:
            raise FileUploadError(f"Failed to refresh presigned URLs: {str(e)}")

    @classmethod
    def _upload_multipart(cls, obj, local_path, local_md5, **kwargs):
        """Enhanced multipart upload with parallel parts and presigned URL refresh"""
        _client = kwargs.get("client") or cls._client or client
        num_processes = kwargs.get("num_processes", 1)
        max_retries = kwargs.get("max_retries", 3)

        try:
            # Get initial presigned URLs
            presigned_urls = obj.presigned_urls
            total_parts = len(presigned_urls)

            print(
                f"Starting multipart upload with {total_parts} parts using {num_processes} worker(s)..."
            )

            # Initialize progress tracker
            progress_tracker = UploadProgressTracker(total_parts, obj.size)

            # Prepare part upload tasks
            part_tasks = []
            for i, part_info in enumerate(presigned_urls):
                part_tasks.append(
                    {
                        "part_number": part_info.part_number,
                        "start_byte": part_info.start_byte,
                        "end_byte": part_info.end_byte,
                        "part_size": part_info.size,
                        "upload_url": part_info.upload_url,
                        "part_index": i,
                        "max_retries": max_retries,
                        "upload_id": obj.upload_id,
                        "upload_key": obj.upload_key,
                    }
                )

            # Upload parts in parallel or sequential
            if num_processes > 1:
                parts = cls._upload_parts_parallel(
                    local_path,
                    part_tasks,
                    obj,
                    _client,
                    num_processes,
                    progress_tracker,
                )
            else:
                parts = cls._upload_parts_sequential(
                    local_path, part_tasks, obj, _client, progress_tracker
                )

            # Complete multipart upload
            return cls._complete_multipart_upload(obj, parts, _client, local_path)

        except Exception as e:
            obj.delete(force=True)
            raise FileUploadError("Multipart upload failed: {}".format(str(e)))

    @classmethod
    def _upload_parts_parallel(
        cls, local_path, part_tasks, obj, _client, num_processes, progress_tracker
    ):
        """Upload parts in parallel using ThreadPoolExecutor"""

        parts = [None] * len(part_tasks)  # Pre-allocate array for ordered results

        failed_parts = cls._upload_parts_with_threadpool(
            local_path,
            part_tasks,
            obj,
            _client,
            parts,
            progress_tracker,
            num_processes,
        )

        # Retry failed parts
        if failed_parts:
            print(f"Retrying {len(failed_parts)} failed parts in parallel...")
            for task in failed_parts:
                task["max_retries"] = 3
            retry_failed = cls._upload_parts_with_threadpool(
                local_path,
                failed_parts,
                obj,
                _client,
                parts,
                progress_tracker,
                num_processes,
            )
            if retry_failed:
                raise Exception(
                    f"Failed to upload {len(retry_failed)} parts after retry"
                )

        return [part for part in parts if part is not None]

    @classmethod
    def _upload_parts_with_threadpool(
        cls,
        local_path,
        part_tasks,
        obj,
        _client,
        parts,
        progress_tracker,
        num_processes,
    ):
        """Common method for uploading parts with ThreadPoolExecutor"""
        from concurrent.futures import ThreadPoolExecutor, as_completed

        failed_parts = []

        with ThreadPoolExecutor(max_workers=num_processes) as executor:
            # Submit all part upload tasks with worker assignment
            future_to_part = {}
            for i, task in enumerate(part_tasks):
                # Assign worker ID to task
                task["worker_id"] = f"Worker-{i % num_processes + 1}"
                future_to_part[
                    executor.submit(
                        cls._upload_single_part,
                        local_path,
                        task,
                        obj,
                        _client,
                        progress_tracker,
                    )
                ] = task

            # Collect results as they complete
            for future in as_completed(future_to_part):
                task = future_to_part[future]
                try:
                    part_result = future.result()
                    parts[task["part_index"]] = part_result
                    # Update progress with part size
                    progress_tracker.update_progress(task["part_size"])
                except Exception as e:
                    # Notify progress tracker about error and print error message
                    progress_tracker.notify_error()
                    print(
                        f"ERROR: {task['worker_id']} failed part {task['part_number']}: {e}"
                    )
                    failed_parts.append(task)

        return failed_parts

    @classmethod
    def _upload_parts_sequential(
        cls, local_path, part_tasks, obj, _client, progress_tracker
    ):
        """Upload parts sequentially with retry logic for failed parts"""
        parts = [None] * len(part_tasks)  # Pre-allocate for consistency
        failed_parts = []

        with open(local_path, "rb") as f:
            for i, task in enumerate(part_tasks):
                try:
                    part_result = cls._upload_single_part(
                        local_path, task, obj, _client, progress_tracker, file_handle=f
                    )
                    parts[i] = part_result
                    # Update progress with part size
                    progress_tracker.update_progress(task["part_size"])
                except Exception as e:
                    # Notify progress tracker about error and print error message
                    progress_tracker.notify_error()
                    print(
                        f"ERROR: Sequential worker failed part {task['part_number']}: {e}"
                    )
                    failed_parts.append(task)

        # Retry failed parts sequentially
        if failed_parts:
            print(f"Retrying {len(failed_parts)} failed parts sequentially...")
            for task in failed_parts:
                try:
                    part_result = cls._upload_single_part(
                        local_path, task, obj, _client, progress_tracker
                    )
                    parts[task["part_index"]] = part_result
                    # Update progress with part size
                    progress_tracker.update_progress(task["part_size"])
                except Exception as e:
                    # Notify progress tracker about error and print error message
                    progress_tracker.notify_error()
                    print(
                        f"FINAL ERROR: Sequential worker part {task['part_number']} failed after all retries: {e}"
                    )
                    raise e

        return [part for part in parts if part is not None]

    @classmethod
    def _upload_single_part(
        cls, local_path, task, obj, _client, progress_tracker=None, file_handle=None
    ):
        """Upload a single part with retry logic and presigned URL refresh"""
        part_number = task["part_number"]
        start_byte = task["start_byte"]
        part_size = task["part_size"]
        max_retries = task["max_retries"]
        upload_id = task["upload_id"]
        upload_key = task["upload_key"]
        worker_id = task.get("worker_id", "Sequential worker")

        for attempt in range(max_retries):
            try:
                # Get fresh presigned URL if this is a retry
                if attempt > 0:
                    print(
                        f"{worker_id} retrying part {part_number} (attempt {attempt + 1}/{max_retries})"
                    )
                    # Refresh presigned URLs for this specific part
                    fresh_urls = cls.refresh_presigned_urls(
                        upload_id=upload_id,
                        key=upload_key,
                        total_size=obj.size,
                        part_numbers=[part_number],
                        client=_client,
                    )
                    upload_url = fresh_urls[0]["upload_url"]
                else:
                    upload_url = task["upload_url"]

                # Read part data
                if file_handle:
                    # Use provided file handle (sequential mode)
                    file_handle.seek(start_byte)
                    chunk_data = file_handle.read(part_size)
                else:
                    # Open file for this part (parallel mode)
                    with open(local_path, "rb") as f:
                        f.seek(start_byte)
                        chunk_data = f.read(part_size)

                if not chunk_data:
                    break

                # Upload without requests-level retry (let our custom retry handle it)
                session = requests.Session()

                headers = {"Content-Length": str(len(chunk_data))}

                # Calculate timeout based on part size
                part_size_mb = len(chunk_data) / (1024 * 1024)
                # Timeout scaling for large parts
                # Formula: 20min base + 30s per MB to handle very large parts
                # This ensures adequate timeout even with slow connections
                base_timeout = 1200  # 20 minutes base
                scaling_factor = 30  # 30 seconds per 1MB
                total_timeout = base_timeout + part_size_mb * scaling_factor

                upload_resp = session.put(
                    upload_url,
                    data=chunk_data,
                    headers=headers,
                    timeout=total_timeout,
                )

                if upload_resp.status_code == 200:
                    etag = upload_resp.headers.get("ETag", "").strip('"')
                    return {"part_number": part_number, "etag": etag}
                else:
                    raise FileUploadError(
                        f"{worker_id} failed part {part_number}: {upload_resp.status_code} - {upload_resp.content}"
                    )

            except Exception as e:
                if attempt == max_retries - 1:  # Last attempt
                    raise FileUploadError(
                        f"{worker_id} part {part_number} failed after {max_retries} attempts: {e}"
                    )

                # Wait before retry (exponential backoff)
                wait_time = 2**attempt
                if progress_tracker:
                    progress_tracker.notify_error()
                print(
                    f"{worker_id} part {part_number} failed \
                    (attempt {attempt + 1}/{max_retries}): {str(e)}, retrying in {wait_time}s..."
                )
                time.sleep(wait_time)

    @classmethod
    def _complete_multipart_upload(cls, obj, parts, _client, local_path):
        """Complete the multipart upload"""
        print("Completing multipart upload...")
        complete_data = {
            "upload_id": obj.upload_id,
            "physical_object_id": obj.upload_key,
            "parts": parts,
        }
        complete_resp = _client.post("/v2/complete_multi_part", complete_data)

        if "message" in complete_resp:
            print(
                f"Successfully uploaded {local_path} to {obj.path} with multipart upload using {len(parts)} parts."
            )
            return obj
        else:
            raise Exception(complete_resp)

    def _object_list_helper(self, **params):
        """Helper method to get objects within"""

        if not self.is_folder:
            raise QuartzBioError(
                "Only folders contain child objects. This is a {}".format(
                    self.object_type
                )
            )

        params["vault_id"] = self.vault_id
        if "recursive" in params:
            params["ancestor_id"] = self.id
            params["limit"] = 1000
        else:
            params["parent_object_id"] = self.id

        items = self.all(client=self._client, **params)
        return items

    def files(self, **params):
        return self._object_list_helper(object_type="file", **params)

    def folders(self, **params):
        return self._object_list_helper(object_type="folder", **params)

    def datasets(self, **params):
        return self._object_list_helper(object_type="dataset", **params)

    def objects(self, **params):
        return self._object_list_helper(**params)

    def ls(self, **params):
        return self.objects(**params)

    def __getattr__(self, name):
        """Shortcut to access attributes of the underlying Dataset resource"""
        from quartzbio.resource import Dataset

        # Valid override attributes that let Object act like a Dataset
        valid_dataset_attrs = [
            # query data
            "lookup",
            "beacon",
            # transform data
            "import_file",
            "export",
            "migrate",
            # dataset meta
            "fields",
            "template",
            "imports",
            "commits",
            # helpers
            "activity",
            "saved_queries",
        ]

        try:
            return self[name]
        except KeyError as err:
            if name in valid_dataset_attrs and self["object_type"] == "dataset":
                return getattr(Dataset(self["id"], client=self._client), name)

            raise AttributeError(*err.args)

    @property
    def dataset(self):
        """Returns the dataset object"""
        from quartzbio import Dataset

        if not self.is_dataset:
            raise QuartzBioError(
                "Only dataset objects have a Dataset resource. This is a {}".format(
                    self.object_type
                )
            )

        return Dataset.retrieve(self["id"], client=self._client)

    @property
    def parent(self):
        """Returns the parent object"""
        if self["parent_object_id"]:
            return self.retrieve(self["parent_object_id"], client=self._client)

        return None

    @property
    def vault(self):
        """Returns the vault object"""
        from . import Vault

        return Vault.retrieve(self["vault_id"], client=self._client)

    @property
    def is_dataset(self):
        return self.object_type == "dataset"

    @property
    def is_folder(self):
        return self.object_type == "folder"

    @property
    def is_file(self):
        return self.object_type == "file"

    @property
    def is_shortcut(self):
        return self.object_type == "shortcut"

    def get_target(self, return_none_target=True):
        if not self.is_shortcut:
            raise QuartzBioError(
                "Only shortcut objects have a target resource. This is a {}".format(
                    self.object_type
                )
            )
        target = self["target"]
        if not target:
            if return_none_target:
                return None
            else:
                raise QuartzBioError("Shortcut target not found.")

        try:
            if target["object_type"] == "url":
                return target["url"]
            elif target["object_type"] == "vault":
                from . import Vault

                return Vault.retrieve(target["id"], client=self._client)
            else:
                return Object.retrieve(target["id"], client=self._client)
        except QuartzBioError as e:
            if e.status_code == 404:
                raise NotFoundError
            else:
                raise e

    @property
    def data_url(self):
        return "/v2/objects/{}/data".format(self.quartzbio_id)

    def has_tag(self, tag):
        """Return True if object contains tag"""

        def lowercase(x):
            return x.lower()

        return lowercase(str(tag)) in map(lowercase, self.tags)

    def tag(self, tags, remove=False, dry_run=False, apply_save=True):
        """Add or remove tags on an object"""

        def is_iterable_non_string(arg):
            """python2/python3 compatible way to check if arg is an iterable but not string"""

            return isinstance(arg, Iterable) and not isinstance(arg, (str, bytes))

        if not is_iterable_non_string(tags):
            tags = [str(tags)]
        else:
            tags = [str(tag) for tag in tags]

        if remove:
            removal_tags = [tag for tag in tags if self.has_tag(tag)]
            if removal_tags:
                print(
                    "{}Notice: Removing tags: {} from object: {}".format(
                        "[Dry Run] " if dry_run else "",
                        ", ".join(removal_tags),
                        self.full_path,
                    )
                )

                tags_for_removal = [tag for tag in tags if self.has_tag(tag)]
                updated_tags = [tag for tag in self.tags if tag not in tags_for_removal]
            else:
                print(
                    "{}Notice: Object {} does not contain any of the "
                    "following tags: {}".format(
                        "[Dry Run] " if dry_run else "", self.full_path, ", ".join(tags)
                    )
                )
                return False
        else:
            new_tags = [tag for tag in tags if not self.has_tag(tag)]
            if new_tags:
                print(
                    "{}Notice: Adding tags: {} to object: {}".format(
                        "[Dry Run] " if dry_run else "",
                        ", ".join(new_tags),
                        self.full_path,
                    )
                )
                updated_tags = self.tags + new_tags
            else:
                print(
                    "{}Notice: Object {} already contains these tags: {}".format(
                        "[Dry Run] " if dry_run else "", self.full_path, ", ".join(tags)
                    )
                )
                return False

        if not dry_run and apply_save:
            self.tags = updated_tags
            self.save()

        return True

    def untag(self, tags, dry_run=False, apply_save=True):
        """Remove tags on an object"""

        return self.tag(tags=tags, remove=True, dry_run=dry_run, apply_save=apply_save)

    def query(self, **params):
        """
        Return the Query or QueryFile object depending on object type
        that represents query results against an object.
        """
        from quartzbio.resource.dataset import Dataset
        from quartzbio.query import QueryFile

        if self.is_dataset:
            return Dataset(self["id"], client=self._client).query(**params)
        elif self.is_file:
            return QueryFile(self["id"], client=self._client, **params)
        else:
            raise QuartzBioError(
                "The functionality is only supported for files and datasets. "
                "This is a {}.".format(self.object_type)
            )

    def archive(self, storage_class=None, follow=False):
        """
        Archive this dataset
        """
        if not self.is_dataset:
            raise QuartzBioError("Only dataset objects can be archived.")

        # The default archive storage class is called "Archive"
        if not storage_class:
            storage_class = "Archive"

        # Updating storage class is only available at the objects endpoint
        self.storage_class = storage_class
        self.save()

        if follow:
            self.dataset.activity(follow=True)

    def restore(self, storage_class=None, follow=False, **kwargs):
        """
        Restore this dataset
        """
        if not self.is_dataset:
            raise QuartzBioError("Only datasets can be restored.")

        # By default, datasets will be restored to Standard-IA
        if not storage_class:
            storage_class = "Standard-IA"

        # Updating storage class is only available at the objects endpoint
        self.storage_class = storage_class
        self.save()

        if follow:
            self.dataset.activity(follow=True)

    def enable_global_beacon(self):
        """
        Enable Global Beacon for this object (datasets only).
        """
        if not self.is_dataset:
            raise QuartzBioError("Only dataset objects can be Global Beacons.")

        return self._client.post(self.instance_url() + "/beacon", {})

    def disable_global_beacon(self):
        """
        Disable Global Beacon for this object (datasets only).
        """
        if not self.is_dataset:
            raise QuartzBioError("Only dataset objects can be Global Beacons.")

        return self._client.delete(self.instance_url() + "/beacon", {})

    def get_global_beacon_status(self, raise_on_disabled=False):
        """
        Retrieves the Global Beacon status for this object (datasets only).
        """
        if not self.is_dataset:
            raise QuartzBioError("Only dataset objects can be Global Beacons.")

        try:
            return self._client.get(self.instance_url() + "/beacon", {})
        except QuartzBioError:
            if raise_on_disabled:
                raise
            return None

    def list_versions(self, include_deleted=True):
        """
        Returns all the versions for this object.
        Only file objects can have versions.
        """
        if not self.is_file:
            raise QuartzBioError("Only file objects can have versions.")

        _client = self._client
        url = self.instance_url() + "/versions"
        params = {"show_deleted": include_deleted}
        response = _client.get(url, params)
        results = convert_to_quartzbio_object(response, client=_client)
        # set up tabulate:
        if len(results.data) > 0:
            list_fields = getattr(results.data[0], "LIST_FIELDS", None)
            if list_fields:
                fields, headers = list(zip(*list_fields))
                results.set_tabulate(fields, headers=headers, sort=False)
        return results

    def delete_version(self, version_id):
        """
        Marks the specified version as deleted.
        Only file objects can have versions.
        """
        if not self.is_file:
            raise QuartzBioError("Only file objects can have versions.")

        url = self.instance_url() + "/versions/{}".format(version_id)
        return self._client.delete(url, {})

    def undelete_version(self, version_id):
        """
        Marks the specified version as not deleted.
        Only file objects can have versions.
        """
        if not self.is_file:
            raise QuartzBioError("Only file objects can have versions.")

        url = self.instance_url() + "/versions/{}".format(version_id)
        return self._client.request("PUT", url)

    def restore_version(self, version_id):
        """
        Sets the current version to the specified version.
        This is done by creating a new version with the same content as the specified version.
        Only file objects can have versions.
        """
        if not self.is_file:
            raise QuartzBioError("Only file objects can have versions.")

        url = self.instance_url() + "/versions/current"
        data = {"version_id": version_id}
        return self._client.post(url, data=data)




from .helper import QuartzBioTestCase


class BatchQueryTest(QuartzBioTestCase):
    def setUp(self):
        super(BatchQueryTest, self).setUp()
        self.dataset = self.client.Object.get_by_full_path(self.TEST_DATASET_FULL_PATH)

    def test_invalid_batch_query(self):
        queries = [
            self.dataset.query(limit=1, fields=["bogus_field"]),
            self.dataset.query(limit=10).filter(bogus_id__gt=100000),
        ]

        results = self.client.BatchQuery(queries).execute()
        self.assertEqual(len(results), 2)
        self.assertEqual(results[0]["status_code"], 400)
        self.assertEqual(results[1]["status_code"], 400)

    def test_batch_query(self):
        queries = [
            self.dataset.query(limit=1),
            self.dataset.query(limit=10).filter(mamit_trnadb__gt=1),
            self.dataset.query(limit=100),
        ]
        results = self.client.BatchQuery(queries).execute()
        self.assertEqual(len(results), 3)
        self.assertEqual(len(results[0]["results"]), 1)
        self.assertEqual(len(results[1]["results"]), 10)
        self.assertEqual(len(results[2]["results"]), 100)



"""QuartzBio DatasetField API Resource"""

from .quartzbio_object import convert_to_quartzbio_object
from .apiresource import CreateableAPIResource
from .apiresource import ListableAPIResource
from .apiresource import UpdateableAPIResource
from .apiresource import DeletableAPIResource


class DatasetField(
    CreateableAPIResource,
    ListableAPIResource,
    DeletableAPIResource,
    UpdateableAPIResource,
):
    """
    Each QuartzBio dataset has a different set of fields, some of
    which can be used as filters. Dataset field resources provide
    users with documentation about each field.
    """

    RESOURCE_VERSION = 2

    def facets(self, **params):
        response = self._client.get(self.facets_url, params)
        return convert_to_quartzbio_object(response, client=self._client)

    def help(self):
        return self.facets()



from selenium import webdriver
import multiprocessing
import requests
import time
import unittest
import percy
import sys
import os

from .utils import invincible, wait_for


class IntegrationTests(unittest.TestCase):
    def percy_snapshot(cls, name):
        if (
            "PERCY_PROJECT" in os.environ
            and os.environ["PERCY_PROJECT"] == "quartzbio/contrib/dash"
        ):
            snapshot_name = "{} - Py{}".format(name, sys.version_info.major).replace(
                "/", "-"
            )
            try:
                cls.percy_runner.snapshot(name=snapshot_name)
            except Exception:
                print('Saving "{}" failed'.format(snapshot_name))

    @classmethod
    def setUpClass(cls):
        super(IntegrationTests, cls).setUpClass()
        cls.driver = webdriver.Chrome()

        if (
            "PERCY_PROJECT" in os.environ
            and os.environ["PERCY_PROJECT"] == "quartzbio/contrib/dash"
        ):
            loader = percy.ResourceLoader(webdriver=cls.driver)
            cls.percy_runner = percy.Runner(loader=loader)

            cls.percy_runner.initialize_build()

    @classmethod
    def tearDownClass(cls):
        super(IntegrationTests, cls).tearDownClass()
        cls.driver.quit()
        if (
            "PERCY_PROJECT" in os.environ
            and os.environ["PERCY_PROJECT"] == "quartzbio/contrib/dash"
        ):
            cls.percy_runner.finalize_build()

    def setUp(self):
        super(IntegrationTests, self).setUp()
        self.driver = webdriver.Chrome()

        def wait_for_element_by_id(id):
            wait_for(
                lambda: None
                is not invincible(lambda: self.driver.find_element_by_id(id))
            )
            return self.driver.find_element_by_id(id)

        self.wait_for_element_by_id = wait_for_element_by_id

        def wait_for_element_by_css_selector(css_selector):
            wait_for(
                lambda: None
                is not invincible(
                    lambda: self.driver.find_element_by_css_selector(css_selector)
                )
            )
            return self.driver.find_element_by_css_selector(css_selector)

        self.wait_for_element_by_css_selector = wait_for_element_by_css_selector

    def tearDown(self):
        super(IntegrationTests, self).tearDown()
        time.sleep(5)
        self.server_process.terminate()
        time.sleep(5)
        self.driver.quit()

    def startServer(self, app):
        def run():
            app.scripts.config.serve_locally = True
            app.run_server(port=8050, debug=False, processes=2)

        # Run on a separate process so that it doesn't block
        self.server_process = multiprocessing.Process(target=run)
        self.server_process.start()
        time.sleep(15)

        # Visit the dash page
        try:
            self.driver.get(
                "http://localhost:8050{}".format(app.config["routes_pathname_prefix"])
            )
        except:
            print("Failed attempt to load page, trying again")
            print(self.server_process)
            print(self.server_process.is_alive())
            time.sleep(5)
            print(requests.get("http://localhost:8050"))
            self.driver.get("http://localhost:8050")

        time.sleep(0.5)

        # Inject an error and warning logger
        logger = """
        window.tests = {};
        window.tests.console = {error: [], warn: [], log: []};

        var _log = console.log;
        var _warn = console.warn;
        var _error = console.error;

        console.log = function() {
            window.tests.console.log.push({
                method: 'log',
                arguments: arguments
            });
            return _log.apply(console, arguments);
        };

        console.warn = function() {
            window.tests.console.warn.push({
                method: 'warn',
                arguments: arguments
            });
            return _warn.apply(console, arguments);
        };

        console.error = function() {
            window.tests.console.error.push({
                method: 'error',
                arguments: arguments
            });
            return _error.apply(console, arguments);
        };
        """
        self.driver.execute_script(logger)



# -*- coding: utf-8 -*-

from quartzbio.resource.quartzbio_object import convert_to_quartzbio_object


class Fake201Response(object):
    status_code = 201
    class_name = None

    def __init__(self, data):
        self.object = {"id": data.get("id", 100), "class_name": self.class_name}

    def json(self):
        return self.object

    def create(self, *args, **kwargs):
        return convert_to_quartzbio_object(self.object)

    def save(self, *args, **kwargs):
        return self.create(*args, **kwargs)

    def retrieve(self, r_id, *args, **kwargs):
        obj = self.create()
        obj.id = r_id
        return obj

    def _retrieve_helper(self, model_name, field_name, error_value, *args, **kwargs):
        obj = self.create()
        for k, v in kwargs.items():
            obj[k] = v

        return obj

    def all(self, *args, **kwargs):
        class ExtendedList(list):
            def quartzbio_objects(self):
                return convert_to_quartzbio_object(self)

        return ExtendedList([self.create()])

    def update_paths(self):
        """Sets vault_name, path and full_path based on mock inputs"""
        if not self.object["vault_name"]:
            self.object["vault_name"] = "mock_vault"

        if not self.object["path"]:
            self.object["path"] = "/{0}".format(
                self.object.get("filename", self.object.get("vault_object_filename"))
            )

        if not self.object["full_path"]:
            self.object["full_path"] = "quartzbio:{0}:{1}".format(
                self.object["vault_name"], self.object["path"]
            )

    def update_dataset(self):
        """Sets dataset_id if object is a dataset"""
        if self.object["object_type"] == "dataset":
            self.object["dataset_id"] = self.object["id"]


class FakeMigrationResponse(Fake201Response):
    class_name = "DatasetMigration"

    def __init__(self, data):
        # Set the default properties of DatasetMigration
        self.object = {
            "class_name": self.class_name,
            "id": 100,
            "commit_mode": "append",
            "source": dict(id=data.get("source_id")),
            "source_params": {
                "fields": None,
                "route": None,
                "exclude_fields": None,
                "filters": None,
                "limit": None,
                "debug": False,
            },
            "target": dict(id=data.get("target_id")),
            "target_fields": [],
            "include_errors": False,
        }
        self.object.update(data)


class FakeExportResponse(Fake201Response):
    class_name = "DatasetExport"

    def __init__(self, data):
        # Set the default properties of DatasetExport
        self.object = {
            "class_name": self.class_name,
            "id": 100,
            "target_fields": [],
            "documents_count": 0,
            "dataset_id": data.get("dataset_id"),
            "dataset": dict(id=data.get("dataset_id")),
            "params": {
                "fields": None,
                "exclude_fields": None,
                "entities": None,
                "genome_build": None,
                "limit": None,
                "filters": None,
                "offset": None,
                "query": None,
                "debug": False,
            },
            "annotator_params": [],
            "metadata": {},
        }
        self.object.update(data)


class FakeVaultResponse(Fake201Response):
    class_name = "Vault"

    def __init__(self, data):
        self.object = {
            "class_name": self.class_name,
            "name": "test_vault",
            "require_unique_paths": True,
            "vault_type": "general",
            "provider": "QuartzBio",
            "id": 100,
            "full_path": "quartzbio:test_vault",
        }
        self.object.update(data)


class FakeObjectResponse(Fake201Response):
    class_name = "Object"

    def __init__(self, data):
        self.object = {
            "class_name": self.class_name,
            "id": 100,
            "parent_object_id": 99,
            "path": None,
            "full_path": None,
            "upload_url": None,
            "size": None,
            "md5": None,
            "filename": "file.json.gz",
            "path": None,
            "full_path": None,
            "vault_name": None,
            "vault_id": None,
            "object_type": "folder",
            "tags": [],
        }
        self.object.update(data)
        self.update_paths()
        self.update_dataset()


class FakeDatasetResponse(Fake201Response):
    class_name = "Dataset"

    def __init__(self, data):
        # Override default object options
        overrides = {
            "object_type": "dataset",
            "filename": data.get("name", "test-dataset"),
            "vault_object_name": data.get("name", "test-dataset"),
            "path": None,
            "full_path": None,
            "class_name": self.class_name,
        }
        data.update(overrides)
        self.object = FakeObjectResponse(data).object
        self.update_paths()


class FakeDatasetTemplateResponse(Fake201Response):
    class_name = "DatasetTemplate"

    def __init__(self, data):
        self.object = {
            "class_name": self.class_name,
            "name": None,
            "id": 100,
            "vault_parent_object_id": 99,
            "account": {
                "name": None,
                "domain": None,
            },
            "fields": None,
            "entity_type": None,
            "validation_params": None,
            "entity_params": None,
            "reader_params": None,
            "annotator_params": None,
        }
        self.object.update(data)


class FakeDatasetTask(Fake201Response):
    class_name = "DatasetTask"

    def __init__(self, data):
        self.object = {
            "class_name": self.class_name,
            "vault_id": None,
            "id": 100,
            "status": "completed",
            "dataset": FakeDatasetResponse(dict(name="fake_dataset")).object,  # noqa
            "account": {
                "name": None,
                "domain": None,
            },
        }
        self.object.update(data)


class FakeDatasetImport(FakeDatasetTask):
    class_name = "DatasetImport"

    def __init__(self, data):
        super(FakeDatasetImport, self).__init__(data)
        self.object["dataset_commits"] = []


def fake_vault_create(*args, **kwargs):
    return FakeVaultResponse(kwargs).create()


def fake_vault_all(*args, **kwargs):
    return FakeVaultResponse(kwargs).all()


def fake_object_all(*args, **kwargs):
    return FakeObjectResponse(kwargs).all()


def fake_object_create(*args, **kwargs):
    return FakeObjectResponse(kwargs).create()


def fake_object_save(*args, **kwargs):
    return FakeObjectResponse(kwargs).save()


def fake_object_retrieve(*args, **kwargs):
    return FakeObjectResponse(kwargs)._retrieve_helper("LogicalObject", *args, **kwargs)


def fake_dataset_create(*args, **kwargs):
    # Dataset.create() is deprecated
    print("WARNING: Your code likely wants to be using FakeObjectCreate.")
    return FakeDatasetResponse(kwargs).create()


def fake_dataset_get_or_create(*args, **kwargs):
    from quartzbio.resource import Object

    dataset_obj = FakeObjectResponse(kwargs).create()

    # Dataset.create requires a full_path, override with this
    full_path, path_dict = Object.validate_full_path(args[0])
    dataset_obj.name = path_dict["filename"]
    dataset_obj.path = path_dict["path"]
    dataset_obj.full_path = full_path
    return FakeDatasetResponse(dict(dataset_obj)).create()


def fake_dataset_import_create(*args, **kwargs):
    return FakeDatasetImport(kwargs).create()


def fake_dataset_tmpl_create(*args, **kwargs):
    return FakeDatasetTemplateResponse(kwargs).create()


def fake_dataset_tmpl_retrieve(rid, *args, **kwargs):
    return FakeDatasetTemplateResponse(kwargs).retrieve(rid)


def fake_migration_create(*args, **kwargs):
    return FakeMigrationResponse(kwargs).create()


def fake_export_create(*args, **kwargs):
    return FakeExportResponse(kwargs).create()



# -*- coding: utf-8 -*-
"""
QuartzBio Python Client
~~~~~~~~~~~~~~~~~~~~~~~

This is the Python client & library for the QuartzBio API.

Have questions or comments? email us at: support@quartzbio.com
"""


__docformat__ = "restructuredtext"

import os as _os
import errno
import logging as _logging
from typing import Literal

# Capture warnings (specifically from urllib3)
try:
    _logging.captureWarnings(True)
except:
    # Python 2.6 doesn't support this
    pass


def _init_logging():
    loglevel_base = _os.environ.get("QUARTZBIO_LOGLEVEL", None) or _os.environ.get(
        "QUARTZBIO_LOGLEVEL", "WARN"
    )
    loglevel_stream = loglevel_base
    logfile = (
        _os.environ.get("QUARTZBIO_LOGFILE", None)
        or _os.environ.get("QUARTZBIO_LOGFILE")
        or "~/.quartzbio/quartzbio.log"
    )
    loglevel_file = loglevel_base

    base_logger = _logging.getLogger("quartzbio")
    base_logger.setLevel(loglevel_base)

    # clear handlers if any exist
    handlers = base_logger.handlers[:]
    for handler in handlers:
        base_logger.removeHandler(handler)
        handler.close()

    if loglevel_stream:
        stream_handler = _logging.StreamHandler()
        stream_handler.setLevel(loglevel_stream)
        stream_fmt = _logging.Formatter("[QuartzBio] %(message)s")
        stream_handler.setFormatter(stream_fmt)
        base_logger.addHandler(stream_handler)

    if logfile:
        logfile_path = _os.path.expanduser(logfile)
        logdir = _os.path.dirname(logfile_path)

        if not _os.path.isdir(logdir):
            # Handle a race condition here when running
            # multiple services that import this package.
            try:
                _os.makedirs(logdir)
            except OSError as err:
                # Re-raise anything other than 'File exists'.
                if err.errno != errno.EEXIST:
                    raise err

        file_handler = _logging.FileHandler(logfile_path)
        file_handler.setLevel(loglevel_file)
        file_fmt = _logging.Formatter(
            "%(asctime)s - %(name)s - %(levelname)s - %(message)s"
        )
        file_handler.setFormatter(file_fmt)
        base_logger.addHandler(file_handler)

    try:
        base_logger.addHandler(_logging.NullHandler())
    except:
        class NullHandler(_logging.Handler):
            def emit(self, record):
                pass

        base_logger.addHandler(NullHandler())


_init_logging()

"""
This is a cached value of client._host, and is kept in for backwards compatibility
Use this with caution, as you should prefer relying on get_api_host() instead
"""
api_host = None


def _set_cached_api_host(host):
    global api_host
    api_host = host


from .version import VERSION  # noqa
from .errors import QuartzBioError
from .query import Query, BatchQuery, Filter, GenomicFilter
from .global_search import GlobalSearch
from .annotate import Annotator, Expression
from .client import client, QuartzBioClient
from .resource import (
    Application,
    Beacon,
    BeaconSet,
    Dataset,
    DatasetCommit,
    DatasetExport,
    DatasetField,
    DatasetImport,
    DatasetMigration,
    DatasetTemplate,
    DatasetRestoreTask,
    DatasetSnapshotTask,
    Group,
    Manifest,
    Object,
    User,
    Vault,
    VaultSyncTask,
    ObjectCopyTask,
    SavedQuery,
    Task,
)


def login(
    api_host: str = None,
    api_key: str = None,
    access_token: str = None,
    name: str = None,
    version: str = None,
    debug: bool = False,
):
    """
    Function to login to the QuartzBio/EDP API when using EDP in a python script.
    Note that another function is used when CLI command `quartzbio login` is used!
    EDP checks user credentials & host URL from multiple sources, in the following order:

    1) Parameters provided (e.g. the parameters of this function)
    2) Environment variables (if the above parameters weren't provided)
    3) quartzbio credentials file stored in the user's HOME directory
        (if parameters and environment variables weren't found)

    :param api_host: the QuartzBio EDP instance's URL to access.
    :param access_token: your user's access token, which you can generate at the EDP website
        (user menu > `Personal Access Tokens`)
    :param api_key: Your API key. You can use this instead of providing an access token
    :param name: name
    :param version: version

    Example:
        .. code-block:: python

            import quartzbio
            quartzbio.login(
                api_host="https://quartzbio.api.az.aws.quartz.bio",
                api_key=YOUR_API_KEY
            )
    """
    token_type: Literal["Bearer", "Token"] = None
    token: str = None

    if access_token:
        token_type = "Bearer"
        token = access_token
    elif api_key:
        token_type = "Token"
        token = api_key

    if api_host or token or debug:
        client.set_credentials(
            api_host, token, token_type=token_type, raise_on_missing=not debug, debug=debug
        )

    client.set_user_agent(name=name, version=version)


def whoami():
    try:
        user = client.whoami()
    except Exception as e:
        print("{} (code: {})".format(e.message, e.status_code))
    else:
        return user


def get_api_host():
    global api_host
    api_host = client._host

    return client._host


get_api_host()


__all__ = [
    "Annotator",
    "Application",
    "Expression",
    "BatchQuery",
    "Beacon",
    "BeaconSet",
    "Dataset",
    "DatasetCommit",
    "DatasetField",
    "DatasetExport",
    "DatasetImport",
    "DatasetMigration",
    "DatasetTemplate",
    "DatasetRestoreTask",
    "DatasetSnapshotTask",
    "Filter",
    "GenomicFilter",
    "GlobalSearch",
    "Group",
    "Manifest",
    "Object",
    "ObjectCopyTask",
    "Query",
    "SavedQuery",
    "QuartzBioClient",
    "QuartzBioError",
    "VaultSyncTask",
    "Task",
    "Vault",
    "User",
    "VERSION",
    "login",
]



import click
import recipes.sync_recipe_utils as sr
import quartzbio as sb

__version__ = '1.0.0'


@click.group()
@click.option('--access-token', help='Manually provide a QuartzBio Access Token')
@click.option('--api-host', help='Override the default QuartzBio API host')
@click.option('--api-key', help='Manually provide a QuartzBio API key')
@click.pass_context
def sync_recipes(ctx, api_key=None, access_token=None, api_host=None):

    sb.login(api_key=api_key, access_token=access_token, api_host=api_host,
             version=__version__, name="QuartzBio Recipes")
    user = sb.User.retrieve()

    click.echo('Logged-in as: {} ({})'.format(
        user.full_name, user.id))


@sync_recipes.command()
@click.argument('recipes_file', nargs=1)
@click.option('--name', help='Name of the recipe')
@click.option('--all', is_flag=True,
              help='Apply the selected mode to all recipes in YAML file')
@click.pass_context
def sync(ctx, recipes_file=None, all=False, name=None):
    if (all and name):
        ctx.fail("Only one of the options --name or --all should be present!")
    elif not any([all, name]):
        ctx.fail("One of the options --name or --all should be present!")

    try:
        yml_recipes = sr.load_recipes_from_yaml(recipes_file)
        if all:
            for yml_recipe in yml_recipes:
                prompt_sync(yml_recipe)
        else:
            yml_recipe = sr.get_recipe_by_name_from_yml(yml_recipes, name)
            if yml_recipe:
                prompt_sync(yml_recipe)

    except Exception as e:
        ctx.fail(e)
        ctx.fail('Invalid path to YAML file with the recipes. Provide the full path to a '
                 'yaml_with_recipes.yml file containing a description of recipes.')


@sync_recipes.command()
@click.argument('recipes_file', nargs=1, required=False)
@click.option('--name', help='Name of the recipe')
@click.option('--all', is_flag=True,
              help='Apply the selected mode to all recipes in YAML file')
@click.pass_context
def delete(ctx, recipes_file, all=False, name=None):
    if (all and name):
        ctx.fail("Only one of the options --name or --all should be present!")
    elif not any([all, name]):
        ctx.fail("One of the options --name or --all should be present!")

    try:
        yml_recipes = sr.load_recipes_from_yaml(recipes_file)
        if all:
            for yml_recipe in yml_recipes:
                prompt_delete("{} (v{})".format(yml_recipe['name'], yml_recipe['version']))
        else:
            prompt_delete(name)

    except Exception as e:
        ctx.fail(e)
        ctx.fail('Invalid path to YAML file with the recipes. Provide the full path to a '
                 'yaml_with_recipes.yml file containing a description of recipes.')


@sync_recipes.command()
@click.argument('recipes_file', nargs=1)
@click.option('--account-recipes', is_flag=True, help='Export recipes for logged in account')
@click.option('--public-recipes', is_flag=True, help='Export public recipes')
@click.pass_context
def export(ctx, recipes_file, account_recipes, public_recipes):
    if account_recipes:
        user = sb.User.retrieve()
        click.echo("Exporting recipes for account {} to {}."
                   .format(user['account']['id'], recipes_file))
        sr.export_recipes_to_yaml(sr.get_account_recipes(user), recipes_file)
        click.echo("Recipes successfully exported!")
    elif public_recipes:
        click.echo("Exporting all public recipes to {}.".format(recipes_file))
        sr.export_recipes_to_yaml(sr.get_public_recipes(), recipes_file)
        click.echo("Recipes successfully exported!")
    else:
        ctx.fail("Export mode should be used with --account-recipes or --public-recipes!")
    return


def prompt_sync(yml_recipe):
    if sb.DatasetTemplate.all(name="{} (v{})".format(yml_recipe['name'], yml_recipe['version'])):
        if click.confirm("Are you sure you want to sync {} (v{}) recipe?"
                                 .format(yml_recipe['name'], yml_recipe['version'])):
            sr.sync_recipe(yml_recipe)
        else:
            click.echo("Aborted.")
    elif click.confirm("Are you sure you want to create {} (v{}) recipe?"
                               .format(yml_recipe['name'], yml_recipe['version'])):
        sr.create_recipe(yml_recipe)
    else:
        click.echo("Aborted.")


def prompt_delete(name):
    if sb.DatasetTemplate.all(name=name):
        if click.confirm("Are you sure you want to delete {} recipe?".format(name)):
            sr.delete_recipe(name)
        else:
            click.echo("Aborted.")
    else:
        click.echo("Requested recipe {} doesn't exist in EDP!".format(name))


if __name__ == '__main__':
    sync_recipes()




from .helper import QuartzBioTestCase


class LookupTests(QuartzBioTestCase):
    def setUp(self):
        super(LookupTests, self).setUp()
        self.dataset = self.client.Object.get_by_full_path(self.TEST_DATASET_FULL_PATH)

    def test_lookup_error(self):
        # Check that incorrect lookup results in empty list.
        lookup_one = self.dataset.lookup("test")
        self.assertEqual(lookup_one, [])

        lookup_two = self.dataset.lookup("test", "nothing")
        self.assertEqual(lookup_two, [])

    def test_lookup_correct(self):
        # Check that lookup with specific sbid is correct.
        records = list(self.dataset.query(limit=2))
        record_one = records[0]
        record_two = records[1]
        sbid_one = record_one["_id"]
        sbid_two = record_two["_id"]

        lookup_one = self.dataset.lookup(sbid_one)
        self.assertEqual(lookup_one[0], record_one)

        lookup_two = self.dataset.lookup(sbid_two)
        self.assertEqual(lookup_two[0], record_two)

        # Check that combining sbids returns list of correct results.
        joint_lookup = self.dataset.lookup(sbid_one, sbid_two)
        self.assertEqual(joint_lookup[0], record_one)
        self.assertEqual(joint_lookup[1], record_two)



import os


OAUTH_CLIENT_ID = os.environ.get("SOVEBIO_OAUTH_CLIENT_ID")
OAUTH_TOKEN = os.environ.get("SOVEBIO_OAUTH_TOKEN")
OAUTH_USERNAME = os.environ.get("SOVEBIO_OAUTH_USERNAME")
OAUTH_PASSWORD = os.environ.get("SOVEBIO_OAUTH_PASSWORD")
OAUTH_DOMAIN = os.environ.get("QUARTZBIO_OAUTH_DOMAIN", "test")

try:
    from .credentials_local import *  # noqa
except:
    pass



# -*- coding: utf-8 -*-

import concurrent.futures
from concurrent.futures import ThreadPoolExecutor

import os
import re
import sys
import gzip
import json
from fnmatch import fnmatch
from collections import defaultdict
import multiprocessing
import signal

import quartzbio

from quartzbio import Task
from quartzbio import Vault
from quartzbio import Object
from quartzbio import Dataset
from quartzbio import DatasetImport
from quartzbio import DatasetTemplate
from quartzbio import GlobalSearch
from quartzbio.utils.files import check_gzip_path
from quartzbio.utils.md5sum import md5sum
from quartzbio.errors import QuartzBioError
from quartzbio.errors import NotFoundError
from quartzbio.client import QuartzBioClient
from quartzbio.client import client as global_client


def should_exclude(path, exclude_paths, dry_run=False, print_logs=True):
    if not exclude_paths:
        return False

    for exclude_path in exclude_paths:
        if fnmatch(path, exclude_path):
            if print_logs:
                print(
                    "{}WARNING: Excluding path {} (via --exclude {})".format(
                        "[Dry Run] " if dry_run else "", path, exclude_path
                    )
                )
            return True

        # An exclude path may be a directory, strip trailing slash and add /*
        # if not already there.
        if not exclude_path.endswith("/*") and fnmatch(
            path, exclude_path.rstrip("/") + "/*"
        ):
            if print_logs:
                print(
                    "{}WARNING: Excluding path {} (via --exclude {})".format(
                        "[Dry Run] " if dry_run else "",
                        path,
                        exclude_path.rstrip("/") + "/*",
                    )
                )
            return True

    return False


def _check_uploaded_folders(
    base_remote_path, local_start, all_folders, follow_shortcuts=False
):
    """Identifies which remote folders already exist so that we can optimize
    which folders to create in the upload workflow.

    Note, due to the asynchronous nature of GlobalSearch, returned folders
    may sometimes already exist. A subsequent exact check is required before
    uploading (as implemented in `Object.create_folder`).

    Args:
        base_remote_path: Base remote parent folder full path to upload to.
        local_start: The name of the base folder to upload.
        all_folders: A list of remote folders to potentially create in full path format.
        follow_shortcuts: Boolean flag specifying if shortcuts on base_remote_path should be resolved.
    Returns:
        all_folder_parts (set): A unique set of folder parts to create that do
            not already exist.

    """

    def _folder_exists(folder_full_path, remote_folders_existing, follow_shortcuts):
        if follow_shortcuts:
            # When following shortcuts we cannot determine which folders exist based on global search.
            # Each folder has to be checked separately.
            try:
                Object.get_by_full_path(
                    full_path=folder_full_path, follow_shortcuts=True
                )
            except NotFoundError:
                return False
            return True
        else:
            return folder_full_path in remote_folders_existing

    upload_root_path, _ = Object.validate_full_path(
        os.path.join(base_remote_path, local_start)
    )
    results = GlobalSearch().filter(path__prefix=upload_root_path, type="folder")
    remote_folders_existing = set([x.full_path for x in results])
    all_folder_parts = set()
    for vault, remote_folder_path in all_folders:
        subfolders = remote_folder_path.lstrip("/").split("/")
        parent_folder_path = subfolders[0]
        # Split each folder into parts and check if these exist
        # Skip root folder as we don't need to create this
        for folder in subfolders[1:]:
            folder_full_path = os.path.join(parent_folder_path, folder)
            if not _folder_exists(
                folder_full_path, remote_folders_existing, follow_shortcuts
            ):
                all_folder_parts.add(folder_full_path)
            parent_folder_path = folder_full_path

    return all_folder_parts


def _upload_folder(
    domain,
    vault,
    base_remote_path,
    base_local_path,
    local_start,
    exclude_paths=None,
    dry_run=False,
    num_processes=1,
    archive_folder=None,
    follow_shortcuts=False,
    max_retries=3,
):
    all_folders = []
    all_files = []

    # Create the upload root folder if it does not exist on the remote
    try:
        upload_root_path, _ = Object.validate_full_path(
            os.path.join(base_remote_path, local_start)
        )
        Object.get_by_full_path(
            upload_root_path, assert_type="folder", follow_shortcuts=follow_shortcuts
        )
    except NotFoundError:
        base_remote_path, path_dict = Object.validate_full_path(base_remote_path)
        base_folder_path = os.path.join(base_remote_path, local_start)
        if dry_run:
            print("[Dry Run] Creating folder {}".format(base_folder_path))
        else:
            all_folders.append((vault, base_folder_path))

    # Create folders and upload files
    for abs_local_parent_path, folders, files in os.walk(base_local_path):
        # Strips off the local path and adds the parent directory at
        # each phase of the loop
        base_local_path_dirname = os.path.dirname(base_local_path)
        if os.name == "nt":
            base_local_path_dirname = re.escape(os.path.dirname(base_local_path))
        local_parent_path = re.sub(
            "^" + base_local_path_dirname, "", abs_local_parent_path
        ).lstrip("/")

        if should_exclude(abs_local_parent_path, exclude_paths, dry_run=dry_run):
            continue

        remote_folder_full_path = os.path.join(base_remote_path, local_parent_path)

        if os.name == "nt":
            remote_folder_full_path = base_remote_path + "/" + local_parent_path.lstrip("\\").replace("\\", "/")

        # Create folders
        for folder in folders:
            new_folder_path = os.path.join(abs_local_parent_path, folder)
            if should_exclude(new_folder_path, exclude_paths, dry_run=dry_run):
                continue

            remote_path = os.path.join(remote_folder_full_path, folder)
            all_folders.append((vault, remote_path))

        # Upload the files that do not yet exist on the remote
        # Pass global client auth parameters to each worker to avoid using default globals
        client_auth = (
            global_client._host,
            global_client._auth.token,
            global_client._auth.token_type,
        )
        for f in files:
            local_file_path = os.path.join(abs_local_parent_path, f)
            if should_exclude(local_file_path, exclude_paths, dry_run=dry_run):
                continue
            all_files.append(
                (
                    local_file_path,
                    remote_folder_full_path,
                    vault.full_path,
                    dry_run,
                    archive_folder,
                    client_auth,
                    follow_shortcuts,
                    max_retries,
                )
            )

    if num_processes > 1:
        # Only perform optimization if parallelization is requested by the user
        all_folder_parts = _check_uploaded_folders(
            base_remote_path, local_start, all_folders, follow_shortcuts
        )
    else:
        all_folder_parts = set([x[1] for x in all_folders])
    # Create folders serially since these require
    # previous folders to be created so that parent_object_id can
    # be populated
    all_folder_parts = sorted(all_folder_parts, key=lambda x: len(x.split("/")))
    for folder in all_folder_parts:
        if dry_run:
            try:
                f = Object.get_by_full_path(folder)
                if not f.is_folder:
                    print("[Dry Run] {} is not a folder - this will cause an error on upload.".format(folder))
                else:
                    print("[Dry Run] Folder {} already exists - skipping creation".format(folder))
            except NotFoundError:
                print("[Dry Run] Creating folder {}".format(folder))
        else:
            Object.create_folder(vault, folder)

    # Create files in parallel
    # Signal handling allows for graceful exit upon KeyboardInterrupt
    original_sigint_handler = signal.signal(signal.SIGINT, signal.SIG_IGN)
    pool = multiprocessing.Pool(num_processes)
    signal.signal(signal.SIGINT, original_sigint_handler)
    try:
        for result in pool.imap_unordered(_create_file_job, all_files):
            # Check if an exception was raised by the pool worker
            if isinstance(result, Exception):
                raise result
    except KeyboardInterrupt:
        print("Caught KeyboardInterrupt, terminating workers and cancelling upload.")
        pool.terminate()
    else:
        pool.close()


def _create_file_job(args):
    """Uploads a single file from local storage to EDP. Args are
    packed into a single tuple to facilitate multiprocessing.

    Args:
        args[0] (local_file_path): Path to local file.
        args[1] (remote_folder_path): Path to remote parent folder.
        args[2] (vault_path): Path to remote vault.
        args[3] (dry_run): Whether to performa dry run.
        args[4] (archive_folder): An archive folder to move existing files into
        args[5] (client_auth): Tuple containing API host, token, and token type
        args[6] (follow_shortcuts): Boolean to follow shortcuts on the remote_folder_path
        args[7] (max_retries): Maximum number of retries per upload part
    Returns:
        None or Exception if exception is raised.
    """
    try:
        (
            local_file_path,
            remote_folder_full_path,
            vault_path,
            dry_run,
            archive_folder,
            client_auth,
            follow_shortcuts,
            max_retries,
        ) = args
        if dry_run:
            print(
                "[Dry Run] Uploading {} to {}".format(
                    local_file_path, remote_folder_full_path
                )
            )
            return
        # Provides the global host, token, token_type
        client = QuartzBioClient(*client_auth)

        remote_parent = None
        try:
            remote_parent = Object.get_by_full_path(
                remote_folder_full_path,
                assert_type="folder",
                follow_shortcuts=follow_shortcuts,
                client=client
            )
        except NotFoundError as e:
            if not dry_run:
                raise e

        if dry_run:
            if not _object_exists(remote_parent, local_file_path, client):
                print("[Dry Run] Uploading {} to {}".format(
                    local_file_path, remote_folder_full_path))
                return
            else:
                print("[Dry Run] File {} already exists at {} - skipping upload".format(
                    local_file_path, remote_folder_full_path))
                return

        Object.upload_file(
            local_file_path,
            remote_parent.path,
            remote_parent.vault.full_path,
            archive_folder=archive_folder,
            follow_shortcuts=follow_shortcuts,
            num_processes=1,  # Default for single file uploads in parallel processing
            max_retries=max_retries,
            client=client,
        )
        return
    except KeyboardInterrupt as e:
        raise e
    except Exception as e:
        return e


def _object_exists(remote_parent, local_path, _client):
    if remote_parent is None:
        return False
    full_path, path_dict = Object.validate_full_path(os.path.join('{}:{}'.format(
        remote_parent.vault.full_path, remote_parent.path), os.path.basename(local_path)), client=_client)
    try:
        obj = Object.get_by_full_path(full_path, client=_client)
        if not obj.is_file:
            return False
        else:
            # Check if the md5sum matches
            local_md5 = md5sum(local_path)[0]
            remote_md5 = obj.get("md5")
            if remote_md5 and remote_md5 == local_md5:
                return True
            else:
                return False
    except NotFoundError:
        return False


def _create_template_from_file(template_file, dry_run=False):
    mode = "r"
    fopen = open
    if check_gzip_path(template_file):
        mode = "rb"
        fopen = gzip.open

    # Validate the template file
    with fopen(template_file, mode) as fp:
        try:
            template_json = json.load(fp)
        except:
            print(
                "Template file {0} could not be loaded. Please "
                "pass valid JSON".format(template_file)
            )
            sys.exit(1)

    if dry_run:
        template = DatasetTemplate(**template_json)
        print("A new dataset template will be created from: {0}".format(template_file))
    else:
        template = DatasetTemplate.create(**template_json)
        print("A new dataset template was created with id: {0}".format(template.id))

    return template


def create_dataset(args, template=None):
    """
    Attempt to create a new dataset given the following params:

        * template_id
        * template_file
        * capacity
        * tag
        * metadata
        * metadata_json_file
        * create_vault
        * full_path
        * dry_run
    """
    if args.dry_run:
        print("NOTE: Running create-dataset command in dry run mode")

    full_path, path_dict = Object.validate_full_path(args.full_path)

    try:
        # Fail if a dataset already exists.
        Object.get_by_full_path(full_path, assert_type="dataset")
        print("A dataset already exists at path: {0}".format(full_path))
        sys.exit(1)
    except NotFoundError:
        pass

    # Accept a template_id or a template_file
    if template:
        # Template has already been validated/created
        # in the import command that called this
        pass
    elif args.template_id:
        try:
            template = DatasetTemplate.retrieve(args.template_id)
        except QuartzBioError as e:
            if e.status_code != 404:
                raise e
            print("No template with ID {0} found!".format(args.template_id))
            sys.exit(1)
    elif args.template_file:
        template = _create_template_from_file(args.template_file, args.dry_run)
    else:
        template = None

    if template:
        print(
            "Creating new dataset {0} using the template '{1}'.".format(
                full_path, template.name
            )
        )
        fields = template.fields
        description = "Created with dataset template: {0}".format(str(template.id))
    else:
        fields = []
        description = None

    # Create dataset metadata
    # Looks at --metadata_json_file first and will update
    # that with any other key/value pairs passed in to --metadata
    metadata = {}
    if args.metadata and args.metadata_json_file:
        print(
            "WARNING: Received --metadata and --metadata-json-file. "
            "Will update the JSON file values with the --metadata values"
        )

    if args.metadata_json_file:
        with open(args.metadata_json_file, "r") as fp:
            try:
                metadata = json.load(fp)
            except:
                print(
                    "Metadata JSON file {0} could not be loaded. Please "
                    "pass valid JSON".format(args.metadata_json_file)
                )
                sys.exit(1)

    if args.metadata:
        metadata.update(args.metadata)

    if args.dry_run:
        print(
            "Creating new '{}' capacity dataset at {}".format(args.capacity, full_path)
        )
        if description:
            print("Description: {}".format(description))
        if fields:
            print("Fields: {}".format(fields))
        if args.tag:
            print("Tags: {}".format(args.tag))
        if metadata:
            print("Metadata: {}".format(metadata))
        return

    return Dataset.get_or_create_by_full_path(
        full_path,
        capacity=args.capacity,
        fields=fields,
        description=description,
        tags=args.tag or [],
        metadata=metadata,
        create_vault=args.create_vault,
    )


def upload(args):
    """
    Given a folder or file, upload all the folders and files contained
    within it, skipping ones that already exist on the remote.
    """

    base_remote_path, path_dict = Object.validate_full_path(args.full_path)

    # Assert the vault exists and is accessible
    vault = Vault.get_by_full_path(path_dict["vault_full_path"])

    follow_shortcuts = True if args.follow_shortcuts else False

    # If not the vault root, validate remote path exists and is a folder
    if path_dict["path"] != "/":
        try:
            Object.get_by_full_path(
                base_remote_path,
                assert_type="folder",
                follow_shortcuts=follow_shortcuts,
            )
        except:
            if not args.create_full_path:
                raise

            if args.dry_run:
                print("[Dry Run] Creating {}".format(base_remote_path))
            else:
                # Create the destination path (including subfolders)
                # if not found
                parent_folder_path = vault.full_path + ":"
                folders = path_dict["path"].lstrip("/").split("/")
                for folder in folders:
                    folder_full_path = os.path.join(parent_folder_path, folder)
                    parent_folder = Object.create_folder(vault, folder_full_path)
                    parent_folder_path = parent_folder.full_path

    # Exit if there are multiple local paths and the
    # exclude paths are not absolute
    base_exclude_paths = args.exclude or []
    if base_exclude_paths and len(args.local_path) > 1:
        rel_exclude_paths = [p for p in base_exclude_paths if not os.path.isabs(p)]
        local_path_parents = set(
            [os.path.dirname(os.path.abspath(p)) for p in args.local_path]
        )
        if rel_exclude_paths and len(local_path_parents) > 1:
            sys.exit(
                "Exiting. Cannot apply the --exclude relative paths when "
                "multiple upload paths with different parent directories "
                "are specified. Make --exclude paths absolute or run "
                "upload paths one at a time."
            )

    for local_path in args.local_path:
        # Expand local path and strip trailing slash
        local_path = os.path.abspath(local_path).rstrip("/")
        local_name = os.path.basename(local_path)

        # add basepath to excludes
        exclude_paths = [
            os.path.join(local_path, os.path.normpath(exclude_path))
            for exclude_path in base_exclude_paths
        ]

        if os.path.isdir(local_path):
            _upload_folder(
                path_dict["domain"],
                vault,
                base_remote_path,
                local_path,
                local_name,
                exclude_paths=exclude_paths,
                dry_run=args.dry_run,
                num_processes=args.num_processes,
                archive_folder=args.archive_folder,
                follow_shortcuts=follow_shortcuts,
                max_retries=args.max_retries,
            )
        else:
            if args.dry_run:
                print(
                    "[Dry Run] Uploading {} to {}".format(local_path, path_dict["path"])
                )
            else:
                Object.upload_file(
                    local_path,
                    path_dict["path"],
                    vault.full_path,
                    archive_folder=args.archive_folder,
                    num_processes=args.num_processes,
                    max_retries=args.max_retries,
                )


def import_file(args):
    """
    Given a dataset and a local path, upload and import the file(s).

    Command arguments (args):

        * create_dataset and it's args
            * capacity
            * template_id
            * template_file
            * capacity
            * tag
            * metadata
            * metadata_json_file
            * create_vault
        * full_path
        * commit_mode
        * remote_source
        * dry_run
        * follow
        * file (list)

    """
    if args.dry_run:
        print("NOTE: Running import command in dry run mode")

    full_path, path_dict = Object.validate_full_path(args.full_path)

    files_list = []
    if args.remote_source:
        # Validate files
        for file_fp in args.file:
            files_ = list(Object.all(glob=file_fp, limit=1000))
            if not files_:
                print(
                    "Did not find any {}files at path {}".format(
                        "remote " if args.remote_source else "", file_fp
                    )
                )
            else:
                for file_ in files_:
                    print("Found file: {}".format(file_.full_path))
                    files_list.append(file_)

    else:
        # Local files
        # Note: if these are globs or folders, then this will
        # create a multi-file manifest which is deprecated
        # and should be updated to one file per import.
        files_list = [fp for fp in args.file]

    if not files_list:
        print(
            "Exiting. No files were found at the following {}paths: {}".format(
                "remote " if args.remote_source else "", ", ".join(args.file)
            )
        )
        sys.exit(1)

    if args.template_id:
        try:
            template = DatasetTemplate.retrieve(args.template_id)
        except QuartzBioError as e:
            if e.status_code != 404:
                raise e
            print("No template with ID {0} found!".format(args.template_id))
            sys.exit(1)
    elif args.template_file:
        template = _create_template_from_file(args.template_file, args.dry_run)
    else:
        template = None

    # Ensure the dataset exists. Create if necessary.
    if args.create_dataset:
        dataset = create_dataset(args, template=template)
    else:
        try:
            dataset = Object.get_by_full_path(full_path, assert_type="dataset")
        except quartzbio.errors.NotFoundError:
            print("Dataset not found: {0}".format(full_path))
            print("Tip: use the --create-dataset flag " "to create one from a template")
            sys.exit(1)

    if args.dry_run:
        print("Importing the following files/paths into dataset: {}".format(full_path))
        for file_ in files_list:
            if args.remote_source:
                print(file_.full_path)
            else:
                print(file_)
        return

    # Generate a manifest from the local files
    imports = []
    for file_ in files_list:
        if args.remote_source:
            kwargs = dict(object_id=file_.id)
        else:
            manifest = quartzbio.Manifest()
            manifest.add(file_)
            kwargs = dict(manifest=manifest.manifest)

        # Add template params
        if template:
            kwargs.update(template.import_params)

        # Create the import
        import_ = DatasetImport.create(
            dataset_id=dataset.id, commit_mode=args.commit_mode, **kwargs
        )

        imports.append(import_)

    if args.follow:
        dataset.activity(follow=True)
    else:
        mesh_url = "https://my.quartzbio.com/activity/"
        print("Your import has been submitted, view details at: {0}".format(mesh_url))

    return imports, dataset


def download(args):
    """
    Given a folder or file, download all the files contained
    within it (not recursive).
    """
    return _download(
        args.full_path,
        args.local_path,
        dry_run=args.dry_run,
        recursive=args.recursive,
        excludes=args.exclude,
        includes=args.include,
        delete=args.delete,
        follow_shortcuts=args.follow_shortcuts,
        num_processes=args.num_processes,
    )


def _download(
    full_path,
    local_folder_path,
    dry_run=False,
    recursive=False,
    excludes=[],
    includes=[],
    delete=False,
    follow_shortcuts=False,
    num_processes=None,
):
    """
    Given a folder or file, download all the files contained
    within it.
    """
    if dry_run:
        print("Running in dry run mode. Not downloading any files.")

    local_folder_path = os.path.expanduser(local_folder_path)
    if not os.path.exists(local_folder_path):
        print("Creating local download folder {}".format(local_folder_path))
        if not dry_run:
            if not os.path.exists(local_folder_path):
                os.makedirs(local_folder_path)

    if recursive:
        _download_recursive(
            full_path,
            local_folder_path,
            dry_run,
            excludes,
            includes,
            delete,
            follow_shortcuts,
            num_processes,
        )
        return

    # API will determine depth based on number of "/" in the glob
    # Add */** to match in any vault (recursive)
    files = Object.all(glob=full_path, limit=1000, object_type="file")
    shortcuts = Object.all(glob=full_path, limit=1000, object_type="shortcut")
    if not files:
        print(
            "No file(s) found at --full-path {}\nIf attempting to download "
            "multiple files, try using a glob 'vault:/path/folder/*'".format(full_path)
        )

    for file_ in files:
        if not dry_run:
            file_.download(local_folder_path)

        print(
            "Downloaded: {} to {}/{}".format(
                file_.full_path, local_folder_path, file_.filename
            )
        )

    if not follow_shortcuts:
        return

    for shortcut_ in shortcuts:
        if not shortcut_["target"]:
            print(
                "None target for shortcut: {} nothing to download.".format(
                    shortcut_.filename
                )
            )
            continue
        if shortcut_["target"]["object_type"] != "file":
            continue

        if not dry_run:
            target = shortcut_.get_target()
            if not target:
                continue
            target.filename = shortcut_.filename
            target.download(local_folder_path)
            print(
                "Downloaded from shortcut: {} to {}/{}".format(
                    shortcut_.full_path, local_folder_path, shortcut_.filename
                )
            )


def _download_recursive(
    full_path,
    local_folder_path,
    dry_run=False,
    excludes=[],
    includes=[],
    delete=False,
    follow_shortcuts=False,
    num_processes=None,
):
    if "**" in full_path:
        raise Exception(
            "Paths containing ** are not compatible with the --recursive flag."
        )

    results = list(
        _resolve_shortcuts_and_get_files(
            full_path=full_path, follow_shortcuts=follow_shortcuts
        )
    )

    num_files = len([x for x in results if x.get("object_type") == "file"])
    print("Found {} files to download.".format(num_files))

    if num_files == 0:
        print("No files found on path.")
        if follow_shortcuts:
            num_shortcuts = len(
                [x for x in results if x.get("object_type") == "shortcut"]
            )
            (
                print(
                    "{} shortcuts found on path. Use --follow-shortcuts flag to download.".format(
                        num_shortcuts
                    )
                )
                if num_shortcuts > 0
                else print("No shortcuts found on path.")
            )
        return

    remote_objects = []
    for file_obj in results:
        depth = len(file_obj.path.split("/"))
        file_obj.depth = depth
        # MD5 not retrieved by GlobalSearch so
        # separate API call is needed
        if file_obj.class_name() == "Vault" or file_obj.is_shortcut:
            continue
        if file_obj.is_file and not file_obj.get("md5"):
            file_object = Object.retrieve(file_obj.id)
            file_object.path = file_obj.path
            file_obj = file_object
        file_obj.depth = depth
        remote_objects.append(file_obj)

    min_depth = min([x.depth for x in remote_objects])
    num_at_min_depth = len([x for x in remote_objects if x.depth == min_depth])
    if num_at_min_depth == 1 and not _is_single_file(remote_objects):
        # when downloading from folder
        base_folder_depth = min_depth
    else:
        # when downloading from vault root or singular file
        base_folder_depth = min_depth - 1

    downloaded_files = set()

    files_to_download = []
    remote_files = [x for x in remote_objects if x.get("object_type") == "file"]
    for remote_file in remote_files:
        rel_parts = remote_file.path.split("/")[base_folder_depth:]
        relative_file_path = os.path.join(*rel_parts)
        local_path = os.path.join(local_folder_path, relative_file_path)

        # Skip over files that are excluded (not recovered by include)
        if should_exclude(
            local_path, excludes, print_logs=False
        ) and not should_exclude(local_path, includes, print_logs=False):
            continue

        # Keep track of remote files to delete locally
        downloaded_files.add(os.path.abspath(local_path))

        # Skip over files that match remote md5 checksum
        if os.path.exists(local_path):
            remote_md5 = remote_file.get("md5")
            if remote_md5 and remote_md5 == md5sum(local_path)[0]:
                print("Skipping {} already in sync".format(local_path))
                continue

        parent_dir = os.path.dirname(local_path)
        print(
            "{}Downloading file {}".format("[Dry run] " if dry_run else "", local_path)
        )
        if not dry_run:
            if not os.path.exists(parent_dir):
                os.makedirs(parent_dir)
            if num_processes is not None:
                file = {"path": local_path, "file": remote_file}
                files_to_download.append(file)
            else:
                remote_file.download(local_path)

    if num_processes is not None:
        _download_in_parallel(files_to_download, num_processes)

    if not delete:
        return
    print("[Warning] Deleting local files not found in remote vault")
    for root, folders, files in os.walk(local_folder_path):
        for file_ in files:
            local_abs_path = os.path.abspath(os.path.join(root, file_))
            if local_abs_path in downloaded_files:
                continue
            print(
                "{}Deleting file {}".format(
                    "[Dry run] " if dry_run else "", local_abs_path
                )
            )
            if not dry_run:
                os.remove(local_abs_path)

        for folder in folders:
            local_abs_path = os.path.abspath(os.path.join(root, folder))
            if len(os.listdir(local_abs_path)) == 0:
                print(
                    "{}Deleting folder {}".format(
                        "[Dry run] " if dry_run else "", local_abs_path
                    )
                )
                os.rmdir(local_abs_path)


def _download_in_parallel(files_to_download, num_processes):
    def _download_worker(file_info):
        try:
            print("downloading to: " + file_info["path"])
            file_info.get("file").download(file_info.get("path"))
        except Exception as e:
            print(
                "Error occurred while downloading file: ({}).".format(
                    file_info.get("path")
                )
            )
            raise e

    if num_processes <= 0:
        num_processes = os.cpu_count()
        print(
            "[Warning] num-processes cannot be less than 1. Defaulting to CPU count: ({})".format(
                num_processes
            )
        )

    print("Downloading in parallel with {} processes.".format(num_processes))

    with ThreadPoolExecutor(max_workers=num_processes) as executor:
        try:
            for result in executor.map(_download_worker, files_to_download):
                pass
        except concurrent.futures.CancelledError as e:
            print("Exception in worker thread:", e)
        except KeyboardInterrupt:
            print("KeyboardInterrupt: Cancelling remaining tasks.")
            executor._threads.clear()
            concurrent.futures.thread._thread_queues.clear()
        except Exception as e:
            print("Exception in worker thread: ", e)


def _resolve_shortcuts_and_get_files(
    full_path, download_path=None, follow_shortcuts=False
):
    full_path, parts = Object.validate_full_path(full_path)
    if not download_path:
        download_path = full_path
    if not download_path.endswith("/"):
        download_path += "/"

    results = GlobalSearch().filter(path__prefix=full_path)

    if full_path.endswith("/*"):
        print(
            "Folder names ending with '/*' are not supported with "
            "the --recursive flag, try again with only the folder name."
        )
        return set()

    if not follow_shortcuts:
        return [r for r in results if r.class_name() != "Vault" and not r.is_shortcut]

    files = []
    for x in results:
        x.path = download_path + _get_relative_download_path(
            base_path=full_path, path_to_object=x.full_path, filename=x.name
        )
        files.append(x)

    shortcuts = [o for o in files if o.class_name() != "Vault" and o.is_shortcut]

    resolved_shortcuts = []
    for shortcut in shortcuts:
        # Global search doesn't return target in the response.
        # Directly accessing target will always return None.
        # Call Object.retrieve() to get the full object.
        shortcut_object = Object.retrieve(shortcut.id)
        try:
            target_object = shortcut_object.get_target()
        except NotFoundError:
            print(
                "[WARNING] The target object for shortcut: ({}) "
                "has been moved, deleted or you don't have permissions to view it.".format(
                    shortcut.full_path
                )
            )
            continue

        if not target_object:
            print(
                "Couldn't find target object for shortcut: {}".format(
                    shortcut.full_path
                )
            )
            continue
        elif shortcut_object.target.object_type == "url":
            print(
                "Found URL shortcut at: ({}) skipping download.".format(
                    shortcut_object.full_path
                )
            )
            continue
        elif shortcut_object.target.object_type == "vault":
            print("Following shortcut to vault: ({})".format(target_object.name))
            resolved_shortcuts += _resolve_shortcuts_and_get_files(
                full_path=target_object.full_path + ":/",
                download_path=shortcut.path + "/" + target_object.name + "/",
                follow_shortcuts=follow_shortcuts,
            )
        elif target_object.is_folder:
            # todo # implement circular shortcut avoiding
            #      # note: it isn't as simple as skipping over already visited folders
            resolved_shortcuts += _resolve_shortcuts_and_get_files(
                full_path=target_object.full_path,
                download_path=shortcut.path + "/",
                follow_shortcuts=follow_shortcuts,
            )
        else:
            # set filename and path for download - keep shortcut structure
            target_object.path = shortcut.path
            resolved_shortcuts.append(target_object)

    return resolved_shortcuts + files


def _get_relative_download_path(base_path, path_to_object, filename):
    relative_path = os.path.relpath(path_to_object, base_path).strip()
    dirname = os.path.dirname(relative_path)
    if dirname:
        return dirname + "/" + filename
    else:
        return filename


def _is_single_file(objects):
    return len(objects) == 1 and objects[0].get("object_type") == "file"


def ls(args):
    """
    Given a QuartzBio remote path, list the files and folders
    """

    if "**" in args.full_path:
        print(
            "Recursive paths containing '**' are not supported by `ls`. "
            "Try the --recursive flag instead."
        )
        return False

    files = _ls(
        args.full_path, recursive=args.recursive, follow_shortcuts=args.follow_shortcuts
    )

    if len(files) == 0:
        print(
            "No file(s) found at '{}'. "
            "Try using glob syntax (vault:/folder/*)".format(args.full_path)
        )
        return False

    return True


def _ls(full_path, recursive=False, follow_shortcuts=False):
    files = list(Object.all(glob=full_path, limit=1000))

    for file_ in files:
        if follow_shortcuts and file_.is_shortcut:
            shortcut = file_.full_path
            try:
                resolved_file = file_.get_target()
                print(
                    "{}  {}  {}  from shortcut: {}".format(
                        resolved_file.last_modified,
                        resolved_file.object_type.ljust(8),
                        resolved_file.full_path.ljust(50),
                        shortcut,
                    )
                )
            except NotFoundError:
                print(
                    "Shortcut {} could not be resolved: "
                    "the target may have been deleted or you may not have permission to access it".format(
                        shortcut
                    )
                )
        else:
            resolved_file = file_
            print(
                "{}  {}  {}".format(
                    resolved_file.last_modified,
                    resolved_file.object_type.ljust(8),
                    resolved_file.full_path,
                )
            )

        if recursive and resolved_file.object_type == "folder":
            _ls(resolved_file.full_path + "/*", recursive=True)

    return files


def should_tag_by_object_type(args, object_):
    """Returns True if object matches object type requirements"""
    valid = True
    if args.tag_folders_only and not object_.is_folder:
        valid = False

    if args.tag_files_only and not object_.is_file:
        valid = False

    if args.tag_datasets_only and not object_.is_dataset:
        valid = False

    if not valid:
        print(
            "{}WARNING: Excluding {} {} by object_type".format(
                "[Dry Run] " if args.dry_run else "",
                object_.object_type,
                object_.full_path,
            )
        )

    return valid


def tag(args):
    """Tags a list of paths with provided tags"""

    objects = []
    for full_path in args.full_path:
        # API will determine depth based on number of "/" in the glob
        # Add */** to match in any vault (recursive)
        objects.extend(list(Object.all(glob=full_path, permission="write", limit=1000)))

    seen_vaults = {}
    taggable_objects = []
    exclusions = args.exclude or []

    # Runs through all objects to get tagging candidates
    # taking exclusions and object_type filters into account.
    for object_ in objects:
        if should_exclude(object_.full_path, exclusions, dry_run=args.dry_run):
            continue

        if should_tag_by_object_type(args, object_):
            taggable_objects.append(object_)
            seen_vaults[object_.vault_id] = 1

    if not taggable_objects:
        print("No taggable objects found at provided locations.")
        return

    # If args.no_input, changes will be applied immediately.
    # Otherwise, prints the objects and if tags will be applied or not
    for object_ in taggable_objects:
        object_.tag(
            args.tag, remove=args.remove, dry_run=args.dry_run, apply_save=args.no_input
        )

    # Prompts for confirmation and then runs with apply_save=True
    if not args.no_input:
        print("")
        res = input(
            "Are you sure you want to apply the above changes to "
            "{} object(s) in {} vault(s)? [y/N] ".format(
                len(taggable_objects), len(seen_vaults.keys())
            )
        )
        print("")
        if res.strip().lower() != "y":
            print("Not applying changes.")
            return

        for object_ in taggable_objects:
            object_.tag(
                args.tag, remove=args.remove, dry_run=args.dry_run, apply_save=True
            )


def show_queue(args):
    """Show running and queued tasks"""
    queue()


def queue(statuses=["running", "queued"]):
    """
    Get all running and queued Tasks for an account
    and groups them by User and status.
    It also prints out the Job queue in the order that they
    will be evaluated.
    """
    task_map = {}
    tasks = Task.all(status=",".join(statuses))
    for task in tasks:
        if task.user.id not in task_map:
            task_map[task.user.id] = []

        task_map[task.user.id].append(task)

    if not task_map:
        print("No running or queued tasks in your account!")
        return

    print("##" * 10)
    print("# Tasks by User")
    print("##" * 10)
    for tasks in task_map.values():
        user = tasks[0].user
        status_map = defaultdict(int)
        print("{} Tasks ({})".format(len(tasks), user.full_name))
        for task in tasks:
            status_map[task.status] += 1

        for status, cnt in status_map.items():
            if cnt:
                print("\t{} are {}".format(cnt, status))



# -*- coding: utf-8 -*-

import os
import glob

from urllib.parse import urlparse

import quartzbio


class Manifest(object):
    """
    Manifests aren't strictly resources, they represent a list
    of remote files (URLs) with additional information that
    can be used for validation (size and MD5).
    """

    manifest = None

    def __init__(self):
        self.manifest = {"files": []}

    def add_file(self, path, **kwargs):
        default_upload_path = quartzbio.Vault.get_or_create_uploads_path()
        vault = quartzbio.Vault.get_personal_vault()
        print("Uploading file: {0} to {1}".format(path, default_upload_path))
        file_ = quartzbio.Object.upload_file(path, default_upload_path, vault.full_path)
        print(
            "Successfuly uploaded file {0} (id:{1} size:{2} md5:{3})".format(
                path, file_.id, file_.size, file_.md5
            )
        )

        self.manifest["files"].append(
            {
                "object_id": file_.id,
                "name": file_.filename,
                "md5": file_.md5,
                "size": file_.size,
                "reader_params": kwargs.get("reader_params"),
                "entity_params": kwargs.get("entity_params"),
                "validation_params": kwargs.get("validation_params"),
            }
        )

    def add_url(self, url, **kwargs):
        manifest_item = dict(url=url, **kwargs)
        self.manifest["files"].append(manifest_item)

    def add(self, *args):
        """
        Add one or more files or URLs to the manifest.
        If files contains a glob, it is expanded.

        All files are uploaded to QuartzBio. The Upload
        object is used to fill the manifest.
        """

        def _is_supported_url(path):
            p = urlparse(path)
            return bool(p.scheme) and p.scheme in ["https", "http"]

        for path in args:
            path = os.path.expanduser(path)
            if _is_supported_url(path):
                self.add_url(path)
            elif os.path.isfile(path):
                self.add_file(path)
            elif os.path.isdir(path):
                for f in os.listdir(path):
                    self.add_file(f)
            elif glob.glob(path):
                for f in glob.glob(path):
                    self.add_file(f)
            else:
                raise ValueError(
                    'Path: "{0}" is not a valid format or does not exist. '
                    "Manifest paths must be local files, directories, blobs "
                    "or URLs with http:// or https://.".format(path)
                )



import os

from .helper import QuartzBioTestCase
from unittest import TestCase
from quartzbio.utils.files import check_gzip_path, separate_filename_extension

FILENAME_PARAMS = [
    {"filename": "test.txt", "base": "test", "ext": ".txt", "compression": ""},
    {
        "filename": "/path/to/test.txt",
        "base": "/path/to/test",
        "ext": ".txt",
        "compression": "",
    },
    {"filename": "test.txt.gz", "base": "test", "ext": ".txt", "compression": ".gz"},
]


class FilenameTests(QuartzBioTestCase, TestCase):
    def test_extract_filename(self):
        for params in FILENAME_PARAMS:
            base, ext, compression = separate_filename_extension(params["filename"])
            self.assertEqual(base, params["base"])
            self.assertEqual(ext, params["ext"])
            self.assertEqual(compression, params["compression"])


class GzipTest(QuartzBioTestCase):
    def test_gzip_file(self):
        path = os.path.join(os.path.dirname(__file__), "data")
        for yes_gzip in ["some_export.json.gz", "sample.vcf.gz"]:
            path = os.path.join(path, yes_gzip)
            self.assertTrue(check_gzip_path(path), path)

        for non_gzip in [
            "sample2.vcf",
            "test_export.json",
            "test_export.csv",
            "test_export.xlsx",
        ]:
            path = os.path.join(path, non_gzip)
            self.assertFalse(check_gzip_path(path), path)



# -*- coding: utf-8 -*-

import quartzbio

from ..client import client
from .credentials import get_credentials
from .credentials import save_credentials
from .credentials import delete_credentials


def login_and_save_credentials(*args):
    """
    CLI command to login and persist credentials to a file
    """
    args = args[0]

    quartzbio.login(
        api_host=args.api_host,
        api_key=args.api_key,
        access_token=args.access_token,
        # name=args.name,
        # version=args.version,
        debug=args.debug,
    )

    # Print information about the current user
    if not client.is_logged_in():
        print("login: client is not logged in!")

        # Verify if user has provided the wrong credentials file
        if client._host:
            if suggested_host := client.validate_host_is_www_url(client._host):
                print(
                    f"Provided API host is: `{client._host}`. "
                    f"Did you perhaps mean `{suggested_host}`?"
                )
        return

    user = client.whoami()
    print_user(user)

    # fixme: how to detect if login was successful
    save_credentials(
        user["email"].lower(),
        client._auth.token,
        client._auth.token_type,
        quartzbio.get_api_host(),
    )
    print("Updated local credentials file.")
    # else:
    #     print(
    #         "You are not logged-in. Visit "
    #         "https://docs.quartzbio.com/#authentication to get started."
    #     )


def logout(*args):
    """
    Delete's the user's locally-stored credentials.
    """
    if get_credentials():
        delete_credentials()
        print("You have been logged out.")
    else:
        print("You are not logged-in.")


def whoami(*args, **kwargs):
    """
    Prints information about the current user.
    Assumes the user is already logged-in.
    """
    try:
        user = client.whoami()
    except Exception as e:
        print("{} (code: {})".format(e.message, e.status_code))
    else:
        print_user(user)


def print_user(user):
    """
    Prints information about the current user.
    """
    email = user["email"]
    domain = user["account"]["domain"]
    print(
        f'You are logged-in to the "{domain}" domain as {email}'
        f' (server: {quartzbio.get_api_host()}).'
    )



"""
helper functions for managing recipes
"""
import sys
from collections import OrderedDict

import quartzbio as sb
import ruamel.yaml as yaml
import click


def create_recipe(description):
    fields = [description['fields']]
    recipe_version = description['version']
    recipe_name = description['name']
    recipe_description = description['description']
    is_public = description['is_public']
    sb.DatasetTemplate.create(
        name="{} (v{})".format(recipe_name, recipe_version),
        description="{}".format(recipe_description) if recipe_description else None,
        template_type="recipe",
        tags=["recipe"],
        is_public=is_public,
        version=recipe_version,
        annotator_params={
            "annotator": "parallel"
        },
        fields=fields
    )


def delete_recipe(recipe_name):
    existing_recipe = sb.DatasetTemplate.all(name=recipe_name)
    if not existing_recipe:
        click.echo("{} doesn't exist!".format(recipe_name))
        return
    for recipe in existing_recipe:
        recipe.delete(force=True)


def sync_recipe(recipe):
    delete_recipe("{} (v{})".format(recipe['name'], recipe['version']))
    create_recipe(recipe)


def get_recipe_by_name_from_yml(all_recipes, name):
    for recipe in all_recipes:
        if recipe["name"] in name and recipe["version"] in name:
            return recipe
    click.echo("{} doesn't exist in the provided YAML file!".format(name))
    return None


def load_recipes_from_yaml(yml_file):
    with open(yml_file, 'r') as yml:
        y = yaml.YAML()
        all_recipes = y.load(yml)
    return all_recipes['recipes']


def get_public_recipes():
    public_recipes = []
    all_templates = sb.DatasetTemplate.all()
    if all_templates:
        for template in all_templates:
            if template['template_type'] == "recipe" and template['is_public']:
                public_recipes.append(template)

    return public_recipes


def get_account_recipes(user):
    account_recipes = []
    all_templates = sb.DatasetTemplate.all()
    if all_templates:
        for template in all_templates:
            if template['template_type'] == "recipe" \
                    and template['account'] == user["account"]["id"]:
                account_recipes.append(template)

    return account_recipes


def export_recipes_to_yaml(recipes, yml_file):
    with open(yml_file, 'w') as outfile:
        class RecipeDumper(yaml.Dumper):
            pass

        class literal(str):
            pass

        def _dict_representer(dumper, data):
            return dumper.represent_mapping(
                yaml.resolver.BaseResolver.DEFAULT_MAPPING_TAG, data.items())

        def _literal_representer(dumper, data):
            return dumper.represent_scalar(
                u'tag:yaml.org,2002:str', data, style='|')

        RecipeDumper.add_representer(dict, _dict_representer)
        RecipeDumper.add_representer(literal, _literal_representer)


        yaml_recipes = []
        for r in recipes:
            recipe_expression = literal(str(dict(r)['fields'][0]['expression']))
            dict(r)['fields'][0]['expression'] = recipe_expression
            recipe_details = {
                "name": dict(r)['name'],
                "description": dict(r)['description'],
                "template_type": "recipe",
                "is_public": "True" if dict(r)['is_public'] else "False",
                "version": dict(r)['version'],
                "annotator_params": {
                    "annotator": "parallel"
                },
                "fields": {i: dict(r)['fields'][0][i]
                           for i in dict(r)['fields'][0]
                           if i not in ['state', 'dictitems'] and dict(r)['fields'][0][i]}
            }
            yaml_recipes.append(recipe_details)
        yaml.dump({"recipes": yaml_recipes}, outfile, RecipeDumper, encoding='utf-8',
                  default_flow_style=False)

    print("Wrote recipe to file: {}".format(yml_file))



# -*- coding: utf-8 -*-

import mock

from .helper import QuartzBioTestCase

from quartzbio.test.client_mocks import fake_migration_create


class TestDatasetMigrations(QuartzBioTestCase):
    @mock.patch("quartzbio.resource.DatasetMigration.create")
    def test_migration_from_query(self, Create):
        Create.side_effect = fake_migration_create

        source = self.client.Dataset(1)
        target = self.client.Dataset(2)

        query = source.query(limit=10, fields=["my_field"]).filter(my_field=999)
        migration = query.migrate(target=target, commit_mode="overwrite", follow=False)
        self.assertEqual(migration.source_id, source.id)
        self.assertEqual(migration.target_id, target.id)
        self.assertEqual(migration.commit_mode, "overwrite")
        self.assertEqual(migration.source_params["filters"], [("my_field", 999)])

    @mock.patch("quartzbio.resource.DatasetMigration.create")
    def test_migration_from_query_target_id(self, Create):
        Create.side_effect = fake_migration_create

        source = self.client.Dataset(1)
        target = self.client.Dataset(2)

        query = source.query(limit=10, fields=["my_field"]).filter(my_field=999)
        migration = query.migrate(
            target=target.id, commit_mode="overwrite", follow=False
        )
        self.assertEqual(migration.source_id, source.id)
        self.assertEqual(migration.target_id, target.id)
        self.assertEqual(migration.commit_mode, "overwrite")
        self.assertEqual(migration.source_params["filters"], [("my_field", 999)])

    @mock.patch("quartzbio.resource.DatasetMigration.create")
    def test_migration_from_query_target_object(self, Create):
        Create.side_effect = fake_migration_create

        source = self.client.Object(1)
        source.dataset_id = source.id
        source.object_type = "dataset"
        target = self.client.Object(2)
        target.object_type = "dataset"

        query = source.query(limit=10, fields=["my_field"]).filter(my_field=999)
        migration = query.migrate(target=target, commit_mode="overwrite", follow=False)
        self.assertEqual(migration.source_id, source.id)
        self.assertEqual(migration.target_id, target.id)
        self.assertEqual(migration.commit_mode, "overwrite")
        self.assertEqual(migration.source_params["filters"], [("my_field", 999)])

    @mock.patch("quartzbio.resource.DatasetMigration.create")
    def test_migration_from_query_target_object_id(self, Create):
        Create.side_effect = fake_migration_create

        source = self.client.Dataset(1)
        target = self.client.Object(2, object_type="dataset")

        query = source.query(limit=10, fields=["my_field"]).filter(my_field=999)
        migration = query.migrate(
            target=target.id, commit_mode="overwrite", follow=False
        )
        self.assertEqual(migration.source_id, source.id)
        self.assertEqual(migration.target_id, target.id)
        self.assertEqual(migration.commit_mode, "overwrite")
        self.assertEqual(migration.source_params["filters"], [("my_field", 999)])

    @mock.patch("quartzbio.resource.DatasetMigration.create")
    def test_migration_target_dataset_id(self, Create):
        Create.side_effect = fake_migration_create

        source = self.client.Dataset(1)
        target = self.client.Dataset(2)
        migration = source.migrate(target=target.id, follow=False)
        self.assertEqual(migration.source_id, source.id)
        self.assertEqual(migration.target_id, target.id)
        self.assertEqual(migration.commit_mode, "append")

    @mock.patch("quartzbio.resource.DatasetMigration.create")
    def test_migration_target_dataset_dataset(self, Create):
        """Target is a Dataset object"""
        Create.side_effect = fake_migration_create

        source = self.client.Dataset(1)
        target = self.client.Dataset(2)
        migration = source.migrate(target=target, follow=False)
        self.assertEqual(migration.source_id, source.id)
        self.assertEqual(migration.target_id, target.id)
        self.assertEqual(migration.commit_mode, "append")

    @mock.patch("quartzbio.resource.DatasetMigration.create")
    def test_migration_target_dataset_object(self, Create):
        """Target is an Object of object_type=dataset"""
        Create.side_effect = fake_migration_create

        source = self.client.Object(1)
        source["object_type"] = "dataset"
        target = self.client.Object(2)
        target["object_type"] = "dataset"
        migration = source.migrate(target=target, follow=False)
        self.assertEqual(migration.source_id, source.id)
        self.assertEqual(migration.target_id, target.id)
        self.assertEqual(migration.commit_mode, "append")

        # Test with source_params
        source_params = {"fields": ["my_field"]}
        migration = source.migrate(
            target=target,
            source_params=source_params,
            commit_mode="overwrite",
            follow=False,
        )
        self.assertEqual(migration.commit_mode, "overwrite")
        self.assertEqual(migration.source_params["fields"], source_params["fields"])
        self.assertEqual(migration.source_id, source.id)
        self.assertEqual(migration.target_id, target.id)



from .apiresource import ListableAPIResource
from .apiresource import DeletableAPIResource
from .apiresource import CreateableAPIResource
from .apiresource import UpdateableAPIResource
from .quartzbio_object import convert_to_quartzbio_object
from .task import Task
from .datasetcommit import follow_commits

import time


class DatasetImport(
    CreateableAPIResource,
    ListableAPIResource,
    UpdateableAPIResource,
    DeletableAPIResource,
):
    """
    DatasetImports represent an import task that takes
    either an object_id or a file manifest (list of file URLs)
    and converts them to a QuartzBio-compatible format which can
    then be indexed by a dataset.

    For interactive use, DatasetImport can be "followed" to watch
    the progression of an import job.
    """

    RESOURCE_VERSION = 2

    LIST_FIELDS = (
        ("id", "ID"),
        ("title", "Title"),
        ("description", "Description"),
        ("status", "Status"),
        ("created_at", "Created"),
    )

    @property
    def dataset(self):
        return convert_to_quartzbio_object(self["dataset"], client=self._client)

    def follow(self, loop=True, sleep_seconds=Task.SLEEP_WAIT_DEFAULT):
        if self.status == "queued":
            print("Waiting for import (id = {0}) to start...".format(self.id))

        import_status = self.status
        while self.status in ["queued", "running"]:
            if self.status != import_status:
                print("Import is now {0} (was {1})".format(self.status, import_status))
                import_status = self.status
                # When status changes to running, indicate
                # pre-processing step.
                if self.status == "running":
                    print(
                        "Processing and validating file(s), "
                        "this may take a few minutes..."
                    )
            elif self.status == "running":
                records_count = self.metadata.get("progress", {}).get(
                    "processed_records", 0
                )
                print(
                    "Import {0} is {1}: {2} records processed".format(
                        self.id, self.status, records_count
                    )
                )
            else:
                print("Import {0} is {1}".format(self.id, self.status))

            if not loop:
                return

            time.sleep(sleep_seconds)
            self.refresh()

        if self.status == "failed":
            print("Import processing and validation failed.")
            print("Reason: {}".format(self.error_message))
            return

        if self.status == "canceled":
            print("Import processing and validation was canceled")
            print("Reason: {}".format(self.error_message))
            return

        # Follow commits until complete
        follow_commits(self, sleep_seconds)

        print(
            "View your imported data: " "https://my.quartzbio.com/data/{0}".format(
                self["dataset"]["id"]
            )
        )



import streamlit as st
import pandas as pd

from dotenv import load_dotenv

# Load environment variables from .env file
load_dotenv()

from quartzbio_streamlit import QuartzBioStreamlit

# from quartzbio.contrib.streamlit.quartzbio_streamlit import QuartzBioStreamlit


@st.cache
def get_personal_vault_items(vault):
    """Getting items from personal vaulet (caching using Streamlit cache mechanism)"""

    files = vault.files()
    folders = vault.folders()
    datasets = vault.datasets()

    return files, folders, datasets


def streamlit_demo_app():
    """Streamlit demo app

    It fetches the QuartzBioClient (initialised with OAuth2 token) from Streamlit session state
    and makes API call through that client.
    """

    # Getting the sovle client from the Streamlit session state
    quartzbio_client = st.session_state.quartzbio_client

    st.title("QuartzBio app")

    # QuartzBio user
    user = quartzbio_client.User.retrieve()
    st.header("{}'s personal vault overview:".format(user["first_name"]))

    # Personal vault
    vault = quartzbio_client.Vault.get_personal_vault()
    st.write(vault["description"])

    files, folders, datasets = get_personal_vault_items(vault)

    # Visualising the stats from personal vault
    data = {
        "Number of files": [files["total"]],
        "Number of datasets": [datasets["total"]],
        "Number of folders": [folders["total"]],
    }
    chart_data = pd.DataFrame.from_dict(
        data, columns=["Number of items"], orient="index"
    )
    st.write(chart_data)
    st.bar_chart(chart_data)

    # Listing items from personal vault
    st.subheader("List selected items")
    option = st.radio("Select QuartzBio platform:", ("Files", "Folders", "Datasets"))
    if option == "Files":
        for item in files:
            st.markdown("- {}".format(item["filename"]))
    elif option == "Folders":
        for item in folders:
            st.markdown("- {}".format(item["filename"]))
    elif option == "Datasets":
        for item in datasets:
            st.markdown("- {}".format(item["filename"]))


# Wrapping Streamlit app with QuartzBio OAuth2
secure_app = QuartzBioStreamlit()
secure_app.wrap(streamlit_app=streamlit_demo_app)




import unittest

import quartzbio
from quartzbio import Query, Filter, GenomicFilter


class FilterTest(unittest.TestCase):
    def test_filter_basic(self):
        f = Filter()
        self.assertEqual(repr(f), "<Filter []>", "empty filter")
        self.assertEqual(repr(~f), "<Filter []>", '"not" of empty filter')

        # Because the order in listing keys is arbitrary, we only
        # test with one entry.
        f1 = Filter(price="Free")
        self.assertEqual(repr(f1), "<Filter [('price', 'Free')]>")
        self.assertEqual(
            repr(~~f1), "<Filter [('price', 'Free')]>", '"not" of empty filter'
        )

        a = quartzbio.query.Filter(chr1="3")
        b = quartzbio.query.Filter(chr2="4")
        self.assertEqual(
            repr(a | b),
            "<Filter [{'or': [('chr1', '3'), ('chr2', '4')]}]>",
            '"or" filter',
        )

        f |= a
        self.assertEqual(repr(f), "<Filter [('chr1', '3')]>", "'or' with empty filter")
        self.assertEqual(
            repr(a), "<Filter [('chr1', '3')]>", "prior 'or' doesn't mung filter"
        )

        filters3 = (
            Filter(omim_id=144650) | Filter(omim_id=144600) | Filter(omim_id=145300)
        )
        self.assertEqual(
            repr(filters3),
            "<Filter [{'or': [('omim_id', 144650),"
            + " ('omim_id', 144600), ('omim_id', 145300)]}]>",
        )

    def test_raw_filters(self):
        # Simple filter
        raw_filter = '[["field_a", "value_a"]]'
        f = Filter(raw_filter)
        self.assertEqual(f.filters, [["field_a", "value_a"]])
        f = Query._process_filters([f])
        self.assertEqual(f, [["field_a", "value_a"]])

        # Compounded filter
        raw_filter = '[["field_a", "value_a"], {"not": {"or": ["field_x", "value_x"]}}]'  # noqa
        f = Filter(raw_filter)
        self.assertEqual(
            f.filters,
            [
                {
                    "and": [
                        ["field_a", "value_a"],
                        {"not": {"or": ["field_x", "value_x"]}},
                    ]
                }
            ],
        )
        f = Query._process_filters([f])
        self.assertEqual(
            f,
            [
                {
                    "and": [
                        ["field_a", "value_a"],
                        {"not": {"or": ["field_x", "value_x"]}},
                    ]
                }
            ],
        )

        # Dict-only as list, and not as a list
        raw_filter = '[{"or": [["field_x", "value_x"]]}]'
        f = Filter(raw_filter)
        self.assertEqual(f.filters, [{"or": [["field_x", "value_x"]]}])
        f = Query._process_filters([f])
        self.assertEqual(f, [{"or": [["field_x", "value_x"]]}])

        raw_filter = '{"or": [["field_x", "value_x"]]}'
        f = Filter(raw_filter)
        self.assertEqual(f.filters, [{"or": [["field_x", "value_x"]]}])
        f = Query._process_filters([f])
        self.assertEqual(f, [{"or": [["field_x", "value_x"]]}])

    def test_combined_raw_filters(self):
        # Combined JSON and regular filter
        raw_filter = '[["field_a", "value_a"]]'
        f = Filter(raw_filter, field_x="value_x")
        self.assertEqual(
            f.filters, [{"and": [("field_x", "value_x"), ["field_a", "value_a"]]}]
        )
        f = Query._process_filters([f])
        self.assertEqual(f, [{"and": [("field_x", "value_x"), ["field_a", "value_a"]]}])

        # Combined, separate raw filters
        raw_filter = '[["field_a", "value_a"]]'
        f = Filter(raw_filter) | Filter(raw_filter)
        self.assertEqual(
            f.filters, [{"or": [["field_a", "value_a"], ["field_a", "value_a"]]}]
        )
        f = Query._process_filters([f])
        self.assertEqual(f, [{"or": [["field_a", "value_a"], ["field_a", "value_a"]]}])

        # Combined list of raw filters
        raw_filter = '[["field_a", "value_a"]]'
        f = Filter(raw_filter, raw_filter)
        self.assertEqual(
            f.filters, [{"and": [["field_a", "value_a"], ["field_a", "value_a"]]}]
        )
        f = Query._process_filters([f])
        self.assertEqual(f, [{"and": [["field_a", "value_a"], ["field_a", "value_a"]]}])

    def test_process_filters(self):
        filters = [("omim_id", None)]
        self.assertEqual(repr(Query._process_filters(filters)), repr(filters))


class GenomicFilterTest(unittest.TestCase):
    def test_single_position(self):
        f = GenomicFilter("chr1", 100)
        expected = [
            "<GenomicFilter [{'and': [('genomic_coordinates.start__lte', 100), ('genomic_coordinates.stop__gte', 100), ('genomic_coordinates.chromosome', '1')]}]>",  # noqa
            "<GenomicFilter [{'and': [('genomic_coordinates.stop__gte', 100), ('genomic_coordinates.start__lte', 100), ('genomic_coordinates.chromosome', '1')]}]>",  # noqa
        ]
        self.assertTrue(repr(f) in expected)

        f = GenomicFilter("chr1", 100, exact=True)
        expected = [
            "<GenomicFilter [{'and': [('genomic_coordinates.stop', 100), ('genomic_coordinates.start', 100), ('genomic_coordinates.chromosome', '1')]}]>",  # noqa
            "<GenomicFilter [{'and': [('genomic_coordinates.start', 100), ('genomic_coordinates.stop', 100), ('genomic_coordinates.chromosome', '1')]}]>",  # noqa
        ]
        self.assertTrue(repr(f) in expected)  # noqa

    def test_range(self):
        f = GenomicFilter("chr1", 100, 200)
        expected = [
            "<GenomicFilter [{'and': [{'or': [{'and': [('genomic_coordinates.start__lte', 100), ('genomic_coordinates.stop__gte', 200)]}, ('genomic_coordinates.start__range', [100, 200]), ('genomic_coordinates.stop__range', [100, 200])]}, ('genomic_coordinates.chromosome', '1')]}]>",  # noqa
            "<GenomicFilter [{'and': [{'or': [{'and': [('genomic_coordinates.stop__gte', 200), ('genomic_coordinates.start__lte', 100)]}, ('genomic_coordinates.start__range', [100, 200]), ('genomic_coordinates.stop__range', [100, 200])]}, ('genomic_coordinates.chromosome', '1')]}]>",  # noqa
        ]
        if repr(f) not in expected:
            print(repr(f))
        self.assertTrue(repr(f) in expected)

        f = GenomicFilter("chr1", 100, 200, exact=True)
        expected = [
            "<GenomicFilter [{'and': [('genomic_coordinates.stop', 200), ('genomic_coordinates.start', 100), ('genomic_coordinates.chromosome', '1')]}]>",  # noqa
            "<GenomicFilter [{'and': [('genomic_coordinates.start', 100), ('genomic_coordinates.stop', 200), ('genomic_coordinates.chromosome', '1')]}]>",  # noqa
        ]
        if repr(f) not in expected:
            print(repr(f))
        self.assertTrue(repr(f) in expected)


if __name__ == "__main__":
    unittest.main()




import os
from typing import Literal, Tuple

from urllib.parse import urlparse

import logging

from requests.auth import AuthBase
import requests

from quartzbio import QuartzBioError
from quartzbio.cli.credentials import get_credentials, netrc

logger = logging.getLogger("quartzbio")


class QuartzBioTokenAuth(AuthBase):
    """Custom auth handler for QuartzBio API token authentication"""

    def __init__(self, token=None, token_type="Token"):
        self.token = token
        self.token_type = token_type

    def __call__(self, r):
        if self.token:
            r.headers["Authorization"] = "{0} {1}".format(self.token_type, self.token)
        return r

    def __repr__(self):
        if self.token:
            return self.token_type
        else:
            return "Anonymous"


def authenticate(
    host: str,
    token: str,
    token_type: Literal["Bearer", "Token"],
    *,
    raise_on_missing: bool = True,
    debug: bool = False
) -> Tuple[str, QuartzBioTokenAuth]:
    """
    Sets login credentials for QuartzBio API authentication.

    :param str host: API host url
    :param str token: API access token or API key
    :param str token_type: API token type. `Bearer` is used for access tokens, while `Token` is used for API Keys.
    :param bool raise_on_missing: Raise an exception if no credentials are available.
    """
    # used for debugging
    source_host = None
    source_token = None

    # Find credentials from environment variables
    if not host:
        host = (
            os.environ.get("QUARTZBIO_API_HOST", None)
            or os.environ.get("EDP_API_HOST", None)
            or os.environ.get("SOLVEBIO_API_HOST", None)
        )

        if host:
            source_host = 'envvars'
    else:
        source_host = 'params'

    if not token:
        api_key = (
            os.environ.get("QUARTZBIO_API_KEY", None)
            or os.environ.get("EDP_API_KEY", None)
            or os.environ.get("SOLVEBIO_API_KEY", None)
        )

        access_token = (
            os.environ.get("QUARTZBIO_ACCESS_TOKEN", None)
            or os.environ.get("EDP_ACCESS_TOKEN", None)
            or os.environ.get("SOLVEBIO_ACCESS_TOKEN", None)
        )

        if access_token:
            token = access_token
            token_type = "Bearer"
        elif api_key:
            token = api_key
            token_type = "Token"

        if token:
            source_token = 'envvars'
    else:
        source_token = 'params'

    # Find credentials from local credentials file
    if not token:
        if creds := get_credentials(host):
            token_type = creds.token_type
            token = creds.token

            if host is None:
                # this happens when user/ennvars provided no API host for the login command
                # but the credentials file still contains login credentials
                host = creds.api_host

                if host:
                    source_host = 'creds'
            if token:
                source_token = 'creds'

    if debug:
        # this will tell the user where QB Client found the credentials from
        creds_path = netrc.path()
        print('\n'.join([
            "Login Debug:",
            f"--> Host: {host}\n    (source: {source_host})",
            f"--> Token Type: {token_type}\n    (source: {source_token})",
            "\n1) source: params",
            "   Means that you've passed this through the login CLI command:",
            "   quartzbio login --host <EDP_HOST> --access_token <EDP_TOKEN>",
            "\n   or the quartzbio.login function:",
            "   import quartzbio",
            "   quartzbio.login(debug=True)",
            "\n2) source: creds",
            "   Means that the QB client has saved your credentials in:",
            f"   {creds_path}",
            "\n3) source: envvars",
            "   Means that you've set your credentials through environment variables:",
            "   QUARTZBIO_API_HOST",
            "   QUARTZBIO_ACCESS_TOKEN",
            "   QUARTZBIO_API_KEY",
        ]))

    if not host:
        if raise_on_missing and not debug:
            raise QuartzBioError("No QuartzBio API host is set")
        else:
            return host, None

    host = validate_api_host_url(host)

    # If the domain ends with .quartzbio.com, determine if
    # we are being redirected. If so, update the url with the new host
    # and log a warning.
    if host and host.rstrip("/").endswith(".api.quartzbio.com"):
        old_host = host.rstrip("/")
        response = requests.head(old_host, allow_redirects=True)
        # Strip the port number from the host for comparison
        new_host = validate_api_host_url(response.url).rstrip("/").replace(":443", "")

        if old_host != new_host:
            logger.warning(
                'API host redirected from "{}" to "{}", '
                "please update your local credentials file".format(old_host, new_host)
            )
            host = new_host

    if token is not None:
        auth = QuartzBioTokenAuth(token, token_type)
    else:
        auth = None

    # TODO: warn user if WWW url is provided in edp_login!

    from quartzbio import _set_cached_api_host
    _set_cached_api_host(host)

    return host, auth


def validate_api_host_url(url):
    """
    Validate QuartzBio API host url.

    Valid urls must not be empty and
    must contain either HTTP or HTTPS scheme.
    """

    # Default to https if no scheme is set
    if "://" not in url:
        url = "https://" + url

    parsed = urlparse(url)
    if parsed.scheme not in ["http", "https"]:
        raise QuartzBioError(
            "Invalid API host: %s. " "Missing url scheme (HTTP or HTTPS)." % url
        )
    elif not parsed.netloc:
        raise QuartzBioError("Invalid API host: %s." % url)

    return parsed.geturl()



# Note that this file is multi-lingual and can be used in both Python
# and POSIX shell.

# This file should define a variable VERSION which we use as the
# debugger version number.
VERSION = "1.3.0rc2"




import json
import flask
import requests
import os

from urllib.parse import urljoin

from dash_auth.auth import Auth

import quartzbio


class QuartzBioAuth(Auth):
    """Handles OAuth2 flows with the QuartzBio API."""

    AUTH_COOKIE_NAME = "dash_quartzbio_auth"
    TOKEN_COOKIE_NAME = "quartzbio_oauth_token"

    DEFAULT_TOKEN_COOKIE_MAX_AGE = 60 * 60 * 24 * 7
    DEFAULT_GRANT_TYPE = "authorization_code"

    OAUTH2_TOKEN_PATH = "/v1/oauth2/token"
    OAUTH2_REVOKE_TOKEN_PATH = "/v1/oauth2/revoke_token"

    def __init__(self, app, app_url, client_id, **kwargs):
        secret_key = kwargs.get("secret_key") or app.server.secret_key

        # Initialize the base Auth class
        super().__init__(app, public_routes=kwargs.get("public_routes", []))

        # Store OAuth configuration
        self._app = app
        self._app_url = app_url
        self._oauth_client_id = client_id
        self._secret_key = secret_key
        self._salt = kwargs.get("salt")

        # Add logout URL
        app.server.add_url_rule(
            "{}_dash-logout".format(app.config["routes_pathname_prefix"]),
            view_func=self.logout,
            methods=["get"],
        )

        self._oauth_redirect_uri = urljoin(
            self._app_url,
            "{}_oauth-redirect".format(self._app.config["requests_pathname_prefix"]),
        )

        # Handle optional parameters
        self._quartzbio_url = kwargs.get("quartzbio_url")
        if not self._quartzbio_url:
            raise Exception(
                'QuartzBioAuth Error: missing required parameter "quartzbio_url"'
            )

        self._api_host = kwargs.get("api_host") or quartzbio.get_api_host()
        self._oauth_client_secret = kwargs.get("client_secret")
        self._oauth_grant_type = kwargs.get("grant_type") or self.DEFAULT_GRANT_TYPE
        self.token_cookie_max_age = kwargs.get(
            "max_age", self.DEFAULT_TOKEN_COOKIE_MAX_AGE
        )

        if self._oauth_grant_type == "implicit":
            self._oauth_response_type = "token"
        elif self._oauth_grant_type == "authorization_code":
            self._oauth_response_type = "code"
        else:
            raise Exception(
                'QuartzBioAuth Error: Unsupported grant type "{}"'.format(
                    self._oauth_grant_type
                )
            )

        _current_path = os.path.dirname(os.path.abspath(__file__))
        with open(os.path.join(_current_path, "oauth-redirect.js"), "r") as f:
            self.oauth_redirect_bundle = f.read()

        with open(os.path.join(_current_path, "login.js"), "r") as f:
            self.login_bundle = f.read()

    def is_authorized(self):
        """Check if the user is authorized by looking for a valid token cookie."""
        try:
            token = flask.request.cookies.get(self.TOKEN_COOKIE_NAME)
            if not token:
                return False

            # Verify the token by making a test API call
            return self.check_view_access(token)
        except:
            return False

    def login_request(self):
        """Return the login page HTML."""
        return self.html(self.login_bundle)

    def auth_wrapper(self, f):
        def wrap(*args, **kwargs):
            if not self.is_authorized():
                # NOTE: We always show the login view instead of
                #       returning 403.
                # TODO - Add a list of valid app paths so we don't
                #        show the login page on asset URLs and Dash endpoints.
                return self.login_request()

            try:
                response = f(*args, **kwargs)
            except Exception as err:
                # Clear the cookie if auth fail
                if getattr(err, "status_code", None) in [401, 403]:
                    response = flask.Response(status=403)
                    self.clear_cookies(response)
                    return response
                else:
                    raise

            # TODO - should set secure in this cookie, not exposed in flask
            # TODO - should set path or domain
            return self.add_access_token_to_response(response)

        # Support for dash-auth >= 1.0.0
        if hasattr(self, "add_access_token_to_response"):
            return wrap
        else:
            return super().auth_wrapper(f)

    def html(self, script):
        return """
            <!doctype html>
            <html lang="en">
            <head>
                <meta charset="UTF-8">
                <title>Log In</title>
            </head>
            <body>
              <div id="react-root"></div>
            </body>
            <script id="_auth-config" type="application/json">
            {}
            </script>
            <script type="text/javascript">{}</script>
            </html>
        """.format(
            json.dumps(
                {
                    "oauth_client_id": self._oauth_client_id,
                    "oauth_response_type": self._oauth_response_type,
                    "quartzbio_url": self._quartzbio_url,
                    "requests_pathname_prefix": self._app.config[
                        "requests_pathname_prefix"
                    ],
                }
            ),
            script,
        )

    def check_view_access(self, oauth_token):
        client = quartzbio.QuartzBioClient(token=oauth_token, token_type="Bearer")
        try:
            client.User.retrieve()
            return True
        except:
            return False

    def login_api(self):
        oauth_data = flask.request.get_json()

        if self._oauth_grant_type == "authorization_code":
            # Authorization Code flow
            auth_code = oauth_data["code"]

            # Request the access token with the auth code
            oauth_data = requests.post(
                urljoin(self._api_host, self.OAUTH2_TOKEN_PATH),
                data={
                    "client_id": self._oauth_client_id,
                    "grant_type": self._oauth_grant_type,
                    "redirect_uri": self._oauth_redirect_uri,
                    "code": auth_code,
                },
            ).json()
        else:
            raise Exception(
                'QuartzBio Auth Error: unsupported grant type "{}"'.format(
                    self._oauth_grant_type
                )
            )

        # Implicit flow returns the access_token in the initial redirect.
        # TODO: Support refresh tokens for authorization code flow.
        oauth_token = oauth_data.get("access_token")
        if not oauth_token:
            # Return the error response to the frontend view
            return flask.Response(
                json.dumps(oauth_data), mimetype="application/json", status=500
            )

        client = quartzbio.QuartzBioClient(token=oauth_token, token_type="Bearer")
        user = client.User.retrieve()
        response = flask.Response(
            json.dumps(user), mimetype="application/json", status=200
        )

        self.set_cookie(
            response=response,
            name=self.TOKEN_COOKIE_NAME,
            value=oauth_token,
            max_age=self.token_cookie_max_age,
            samesite="Lax",
        )

        return response

    def logout(self):
        """Revoke the token and remove the cookie."""
        if self._oauth_client_secret:
            try:
                oauth_token = flask.request.cookies[self.TOKEN_COOKIE_NAME]
                # Revoke the token
                requests.post(
                    urljoin(self._api_host, self.OAUTH2_REVOKE_TOKEN_PATH),
                    data={
                        "client_id": self._oauth_client_id,
                        "client_secret": self._oauth_client_secret,
                        "token": oauth_token,
                    },
                )
            except:
                pass

        response = flask.redirect("/")
        self.clear_cookies(response)
        return response



# -*- coding: utf-8 -*-

from .helper import QuartzBioTestCase


class TestAnnotator(QuartzBioTestCase):
    def test_annotator(self):
        fields = [{"name": "test", "expression": '"hello world"'}]
        records = [{"i": i} for i in range(100)]
        a = self.client.Annotator(fields)

        for i, result in enumerate(a.annotate(records)):
            self.assertEqual(result, {"test": "hello world", "i": i})


class TestExpression(QuartzBioTestCase):
    def test_expression(self):
        answer = self.client.Expression("1+1").evaluate(data_type="integer")
        self.assertEqual(answer, 2)

    def test_expression_with_context(self):
        answer = self.client.Expression("record").evaluate(
            data={"record": 123}, data_type="integer"
        )
        self.assertEqual(answer, 123)



# -*- coding: utf-8 -*-

import time
from mock import patch

import quartzbio.client
from .helper import QuartzBioTestCase


class FakeResponse:
    def __init__(self, headers, status_code):
        self.headers = headers
        self.status_code = status_code

    def json(self):
        return {}


class ClientRateLimit(QuartzBioTestCase):
    """Test of rate-limiting an API request"""

    def setUp(self):
        super(ClientRateLimit, self).setUp()
        self.call_count = 0

    def fake_response(self):
        if self.call_count == 0:
            self.call_count += 1
            return FakeResponse({"retry-after": "0"}, 429)
        else:
            return FakeResponse({}, 200)

    def test_rate_limit(self):
        with patch("quartzbio.client.Session.request") as mock_request:
            mock_request.side_effect = lambda *args, **kw: self.fake_response()
            start_time = time.time()
            dataset = quartzbio.Dataset.retrieve(1)
            elapsed_time = time.time() - start_time
            self.assertTrue(
                isinstance(dataset, quartzbio.Dataset),
                "Got a dataset back (eventually)",
            )
            self.assertTrue(
                elapsed_time > 1.0,
                "Should have delayed for over a second; " "(was %s)" % elapsed_time,
            )
            self.assertEqual(self.call_count, 1)



from ..query import Query

from .apiresource import CreateableAPIResource
from .apiresource import ListableAPIResource
from .apiresource import UpdateableAPIResource
from .apiresource import DeletableAPIResource


class SavedQuery(
    CreateableAPIResource,
    ListableAPIResource,
    UpdateableAPIResource,
    DeletableAPIResource,
):
    """
    A saved query is a set of query parameters that persists,
    giving users the ability to apply them to compatible datasets
    with ease. A dataset is said to be compatible with a saved query
    if it contains all the fields found in said saved query.
    """

    RESOURCE_VERSION = 2

    LIST_FIELDS = (
        ("id", "ID"),
        ("name", "Name"),
        ("is_shared", "Is Shared"),
        ("description", "Description"),
    )

    def query(self, dataset=None):
        if "id" not in self or not self["id"]:
            raise Exception(
                "No SavedQuery ID was provided. "
                "Please instantiate the SavedQuery "
                "object with an ID."
            )
        if not dataset:
            raise Exception(
                "No Dataset was specified. "
                "Please provide either the Dataset object or "
                "the Dataset ID."
            )

        # Can be provided as an object or as an ID.
        try:
            dataset_id = dataset.id
        except AttributeError:
            dataset_id = dataset

        return Query(dataset_id, client=self._client, **self.params)



from .apiresource import ListableAPIResource
from .apiresource import DeletableAPIResource
from .apiresource import CreateableAPIResource
from .quartzbio_object import convert_to_quartzbio_object
from .task import Task
from .datasetcommit import follow_commits

import time


class DatasetMigration(
    CreateableAPIResource, ListableAPIResource, DeletableAPIResource
):
    """
    DatasetMigration represent an task that copies data between
    two Datasets or modifies data within a single Dataset.

    For interactive use, DatasetMigration can be "followed" to watch
    the progression of the task.
    """

    RESOURCE_VERSION = 2

    LIST_FIELDS = (
        ("id", "ID"),
        ("status", "Status"),
        ("source", "Source"),
        ("target", "Target"),
        ("documents_count", "Records"),
        ("created_at", "Created"),
        ("updated_at", "Updated"),
    )

    @property
    def source(self):
        response = self._client.get(self["source"]["url"], {})
        return convert_to_quartzbio_object(response, client=self._client)

    @property
    def target(self):
        response = self._client.get(self["target"]["url"], {})
        return convert_to_quartzbio_object(response, client=self._client)

    def follow(self, loop=True, sleep_seconds=Task.SLEEP_WAIT_DEFAULT):
        _status = self.status
        if self.status == "queued":
            print("Waiting for migration (id = {0}) to start...".format(self.id))

        while self.status in ["queued", "running"]:
            if self.status != _status:
                print("Migration is now {0} (was {1})".format(self.status, _status))
                _status = self.status

            if self.status == "running":
                processed_records = self.metadata.get("progress", {}).get(
                    "processed_records", 0
                )
                print(
                    "Migration '{0}' is {1}: {2}/{3} records migrated".format(
                        self.id, self.status, processed_records, self.documents_count
                    )
                )

            if not loop:
                return

            time.sleep(sleep_seconds)
            self.refresh()

        if self.status == "failed":
            print("Migration failed.")
            print("Reason: {}".format(self.error_message))
            return

        if self.status == "canceled":
            print("Migration was canceled")
            print("Reason: {}".format(self.error_message))
            return

        # Follow commits until complete
        follow_commits(self, sleep_seconds)

        print(
            "View your migrated data: " "https://my.quartzbio.com/data/{0}".format(
                self["target"]["id"]
            )
        )
